import asyncio
from bisect import insort
from collections import defaultdict
from collections import deque
from enum import IntEnum
from enum import IntFlag
from enum import auto
from inspect import isawaitable
from typing import TYPE_CHECKING
from typing import Callable
from typing import Dict
from typing import List

from .exceptions import AttrNotFound
from .i18n import _
from .utils import ensure_iterable

if TYPE_CHECKING:
    from typing import Set


def allways_true(*args, **kwargs):
    return True


class CallbackPriority(IntEnum):
    GENERIC = 0
    INLINE = 10
    DECORATOR = 20
    NAMING = 30
    AFTER = 40


class SpecReference(IntFlag):
    NAME = auto()
    CALLABLE = auto()
    PROPERTY = auto()


SPECS_ALL = SpecReference.NAME | SpecReference.CALLABLE | SpecReference.PROPERTY
SPECS_SAFE = SpecReference.NAME


class CallbackGroup(IntEnum):
    ENTER = auto()
    EXIT = auto()
    VALIDATOR = auto()
    BEFORE = auto()
    ON = auto()
    AFTER = auto()
    COND = auto()

    def build_key(self, specs: "CallbackSpecList") -> str:
        return f"{self.name}@{id(specs)}"


class CallbackSpec:
    """Specs about callbacks.

    At first, `func` can be a name (string), a property or a callable.

    Names, properties and unbounded callables should be resolved to a callable
    before any real call is performed.
    """

    names_not_found: "Set[str] | None" = None
    """List of names that were not found on the model or statemachine"""

    def __init__(
        self,
        func,
        group: CallbackGroup,
        is_convention=False,
        is_event: bool = False,
        cond=None,
        priority: CallbackPriority = CallbackPriority.NAMING,
        expected_value=None,
    ):
        self.func = func
        self.group = group
        self.is_convention = is_convention
        self.is_event = is_event
        self.cond = cond
        self.expected_value = expected_value
        self.priority = priority

        if isinstance(func, property):
            self.reference = SpecReference.PROPERTY
            self.attr_name: str = func and func.fget and func.fget.__name__ or ""
        elif callable(func):
            self.reference = SpecReference.CALLABLE
            self.is_bounded = hasattr(func, "__self__")
            self.attr_name = (
                func.__name__ if not self.is_event or self.is_bounded else f"_{func.__name__}_"
            )
            if not self.is_bounded:
                func.attr_name = self.attr_name
                func.is_event = is_event
        else:
            self.reference = SpecReference.NAME
            self.attr_name = func

        self.may_contain_boolean_expression = (
            not self.is_convention
            and self.group == CallbackGroup.COND
            and self.reference == SpecReference.NAME
        )

    def __repr__(self):
        return f"{type(self).__name__}({self.func!r}, is_convention={self.is_convention!r})"

    def __str__(self):
        name = getattr(self.func, "__name__", self.func)
        if self.expected_value is False:
            name = f"!{name}"
        return name

    def __eq__(self, other):
        return self.func == other.func and self.group == other.group

    def __hash__(self):
        return id(self)


class SpecListGrouper:
    def __init__(self, list: "CallbackSpecList", group: CallbackGroup) -> None:
        self.list = list
        self.group = group
        self.key = group.build_key(list)

    def add(self, callbacks, **kwargs):
        self.list.add(callbacks, group=self.group, **kwargs)
        return self

    def __call__(self, callback):
        return self.list._add_unbounded_callback(callback, group=self.group)

    def _add_unbounded_callback(self, func, is_event=False, transitions=None, **kwargs):
        self.list._add_unbounded_callback(
            func,
            is_event=is_event,
            transitions=transitions,
            group=self.group,
            **kwargs,
        )

    def __iter__(self):
        return (item for item in self.list if item.group == self.group)


class CallbackSpecList:
    """List of {ref}`CallbackSpec` instances"""

    def __init__(self, factory=CallbackSpec):
        self.items: List[CallbackSpec] = []
        self.conventional_specs = set()
        self._groupers: Dict[CallbackGroup, SpecListGrouper] = {}
        self.factory = factory

    def __repr__(self):
        return f"{type(self).__name__}({self.items!r}, factory={self.factory!r})"

    def _add_unbounded_callback(self, func, transitions=None, **kwargs):
        """This list was a target for adding a func using decorator
        `@<state|event>[.on|before|after|enter|exit]` syntax.

        If we assign ``func`` directly as callable on the ``items`` list,
        this will result in an `unbounded method error`, with `func` expecting a parameter
        ``self`` not defined.

        The implemented solution is to resolve the collision giving the func a reference method.
        To update It's callback when the name is resolved on the
        :func:`StateMachineMetaclass.add_from_attributes`.
        If the ``func`` is bounded It will be used directly, if not, it's ref will be replaced
        by the given attr name and on `statemachine._setup()` the dynamic name will be resolved
        properly.

        Args:
            func (callable): The decorated method to add on the transitions occurs.
            is_event (bool): If the func is also an event, we'll create a trigger and link the
                event name to the transitions.
            transitions (TransitionList): If ``is_event``, the transitions to be attached to the
                event.

        """
        self._add(func, **kwargs)
        func._transitions = transitions

        return func

    def __iter__(self):
        return iter(self.items)

    def clear(self):
        self.items = []

    def grouper(self, group: CallbackGroup) -> SpecListGrouper:
        if group not in self._groupers:
            self._groupers[group] = SpecListGrouper(self, group)
        return self._groupers[group]

    def _add(self, func, group: CallbackGroup, **kwargs):
        if isinstance(func, CallbackSpec):
            spec = func
        else:
            spec = self.factory(func, group, **kwargs)

        if spec in self.items:
            return

        self.items.append(spec)
        if spec.is_convention:
            self.conventional_specs.add(spec.func)
        return spec

    def add(self, callbacks, group: CallbackGroup, **kwargs):
        if callbacks is None:
            return self

        unprepared = ensure_iterable(callbacks)
        for func in unprepared:
            self._add(func, group=group, **kwargs)

        return self


class CallbackWrapper:
    def __init__(
        self,
        callback: Callable,
        condition: Callable,
        meta: "CallbackSpec",
        unique_key: str,
    ) -> None:
        self._callback = callback
        self._iscoro = getattr(callback, "is_coroutine", False)
        self.condition = condition
        self.meta = meta
        self.unique_key = unique_key
        self.expected_value = self.meta.expected_value

    def __repr__(self):
        return f"{type(self).__name__}({self.unique_key})"

    def __str__(self):
        return str(self.meta)

    def __lt__(self, other):
        return self.meta.priority < other.meta.priority

    async def __call__(self, *args, **kwargs):
        value = self._callback(*args, **kwargs)
        if isawaitable(value):
            value = await value

        if self.expected_value is not None:
            return bool(value) == self.expected_value
        return value

    def call(self, *args, **kwargs):
        value = self._callback(*args, **kwargs)
        if self.expected_value is not None:
            return bool(value) == self.expected_value
        return value


class CallbacksExecutor:
    """A list of callbacks that can be executed in order."""

    def __init__(self):
        self.items: List[CallbackWrapper] = deque()
        self.items_already_seen = set()

    def __iter__(self):
        return iter(self.items)

    def __repr__(self):
        return f"{type(self).__name__}({self.items!r})"

    def __str__(self):
        return ", ".join(str(c) for c in self)

    def add(self, key: str, spec: CallbackSpec, builder: Callable[[], Callable]):
        if key in self.items_already_seen:
            return

        self.items_already_seen.add(key)

        condition = spec.cond if spec.cond is not None else allways_true
        wrapper = CallbackWrapper(
            callback=builder(),
            condition=condition,
            meta=spec,
            unique_key=key,
        )

        insort(self.items, wrapper)

    async def async_call(self, *args, **kwargs):
        return await asyncio.gather(
            *(
                callback(*args, **kwargs)
                for callback in self
                if callback.condition(*args, **kwargs)
            )
        )

    async def async_all(self, *args, **kwargs):
        coros = [condition(*args, **kwargs) for condition in self]
        for coro in asyncio.as_completed(coros):
            if not await coro:
                return False
        return True

    def call(self, *args, **kwargs):
        return [
            callback.call(*args, **kwargs)
            for callback in self
            if callback.condition(*args, **kwargs)
        ]

    def all(self, *args, **kwargs):
        for condition in self:
            if not condition.call(*args, **kwargs):
                return False
        return True


class CallbacksRegistry:
    def __init__(self) -> None:
        self._registry: Dict[str, CallbacksExecutor] = defaultdict(CallbacksExecutor)
        self.has_async_callbacks: bool = False

    def __getitem__(self, key: str) -> CallbacksExecutor:
        return self._registry[key]

    def check(self, specs: CallbackSpecList):
        for meta in specs:
            if meta.is_convention:
                continue

            if any(
                callback for callback in self[meta.group.build_key(specs)] if callback.meta == meta
            ):
                continue

            if meta.names_not_found:
                raise AttrNotFound(
                    _("Did not found name '{}' from model or statemachine").format(
                        ", ".join(meta.names_not_found)
                    ),
                )
            raise AttrNotFound(
                _("Did not found name '{}' from model or statemachine").format(meta.func)
            )

    def async_or_sync(self):
        self.has_async_callbacks = any(
            callback._iscoro for executor in self._registry.values() for callback in executor
        )

    def call(self, key: str, *args, **kwargs):
        if key not in self._registry:
            return []
        return self._registry[key].call(*args, **kwargs)

    def async_call(self, key: str, *args, **kwargs):
        return self._registry[key].async_call(*args, **kwargs)

    def all(self, key: str, *args, **kwargs):
        if key not in self._registry:
            return True
        return self._registry[key].all(*args, **kwargs)

    def async_all(self, key: str, *args, **kwargs):
        return self._registry[key].async_all(*args, **kwargs)

    def str(self, key: str) -> str:
        if key not in self._registry:
            return ""
        return str(self._registry[key])



"""The ``celery result`` program, used to inspect task results."""
import click

from celery.bin.base import CeleryCommand, CeleryOption, handle_preload_options


@click.command(cls=CeleryCommand)
@click.argument('task_id')
@click.option('-t',
              '--task',
              cls=CeleryOption,
              help_group='Result Options',
              help="Name of task (if custom backend).")
@click.option('--traceback',
              cls=CeleryOption,
              is_flag=True,
              help_group='Result Options',
              help="Show traceback instead.")
@click.pass_context
@handle_preload_options
def result(ctx, task_id, task, traceback):
    """Print the return value for a given task id."""
    app = ctx.obj.app

    result_cls = app.tasks[task].AsyncResult if task else app.AsyncResult
    task_result = result_cls(task_id)
    value = task_result.traceback if traceback else task_result.get()

    # TODO: Prettify result
    ctx.obj.echo(value)



"""Get loader by name.

Loaders define how configuration is read, what happens
when workers start, when tasks are executed and so on.
"""
from celery.utils.imports import import_from_cwd, symbol_by_name

__all__ = ('get_loader_cls',)

LOADER_ALIASES = {
    'app': 'celery.loaders.app:AppLoader',
    'default': 'celery.loaders.default:Loader',
}


def get_loader_cls(loader):
    """Get loader class by name/alias."""
    return symbol_by_name(loader, LOADER_ALIASES, imp=import_from_cwd)



"""System information utilities."""
from __future__ import annotations

import os
from math import ceil

from kombu.utils.objects import cached_property

__all__ = ('load_average', 'df')


if hasattr(os, 'getloadavg'):

    def _load_average() -> tuple[float, ...]:
        return tuple(ceil(l * 1e2) / 1e2 for l in os.getloadavg())

else:  # pragma: no cover
    # Windows doesn't have getloadavg
    def _load_average() -> tuple[float, ...]:
        return 0.0, 0.0, 0.0,


def load_average() -> tuple[float, ...]:
    """Return system load average as a triple."""
    return _load_average()


class df:
    """Disk information."""

    def __init__(self, path: str | bytes | os.PathLike) -> None:
        self.path = path

    @property
    def total_blocks(self) -> float:
        return self.stat.f_blocks * self.stat.f_frsize / 1024

    @property
    def available(self) -> float:
        return self.stat.f_bavail * self.stat.f_frsize / 1024

    @property
    def capacity(self) -> int:
        avail = self.stat.f_bavail
        used = self.stat.f_blocks - self.stat.f_bfree
        return int(ceil(used * 100.0 / (used + avail) + 0.5))

    @cached_property
    def stat(self) -> os.statvfs_result:
        return os.statvfs(os.path.abspath(self.path))



"""Task Routing.

Contains utilities for working with task routers, (:setting:`task_routes`).
"""
import fnmatch
import re
from collections import OrderedDict
from collections.abc import Mapping

from kombu import Queue

from celery.exceptions import QueueNotFound
from celery.utils.collections import lpmerge
from celery.utils.functional import maybe_evaluate, mlazy
from celery.utils.imports import symbol_by_name

try:
    Pattern = re._pattern_type
except AttributeError:  # pragma: no cover
    # for support Python 3.7
    Pattern = re.Pattern

__all__ = ('MapRoute', 'Router', 'expand_router_string', 'prepare')


class MapRoute:
    """Creates a router out of a :class:`dict`."""

    def __init__(self, map):
        map = map.items() if isinstance(map, Mapping) else map
        self.map = {}
        self.patterns = OrderedDict()
        for k, v in map:
            if isinstance(k, Pattern):
                self.patterns[k] = v
            elif '*' in k:
                self.patterns[re.compile(fnmatch.translate(k))] = v
            else:
                self.map[k] = v

    def __call__(self, name, *args, **kwargs):
        try:
            return dict(self.map[name])
        except KeyError:
            pass
        except ValueError:
            return {'queue': self.map[name]}
        for regex, route in self.patterns.items():
            if regex.match(name):
                try:
                    return dict(route)
                except ValueError:
                    return {'queue': route}


class Router:
    """Route tasks based on the :setting:`task_routes` setting."""

    def __init__(self, routes=None, queues=None,
                 create_missing=False, app=None):
        self.app = app
        self.queues = {} if queues is None else queues
        self.routes = [] if routes is None else routes
        self.create_missing = create_missing

    def route(self, options, name, args=(), kwargs=None, task_type=None):
        kwargs = {} if not kwargs else kwargs
        options = self.expand_destination(options)  # expands 'queue'
        if self.routes:
            route = self.lookup_route(name, args, kwargs, options, task_type)
            if route:  # expands 'queue' in route.
                return lpmerge(self.expand_destination(route), options)
        if 'queue' not in options:
            options = lpmerge(self.expand_destination(
                self.app.conf.task_default_queue), options)
        return options

    def expand_destination(self, route):
        # Route can be a queue name: convenient for direct exchanges.
        if isinstance(route, str):
            queue, route = route, {}
        else:
            # can use defaults from configured queue, but override specific
            # things (like the routing_key): great for topic exchanges.
            queue = route.pop('queue', None)

        if queue:
            if isinstance(queue, Queue):
                route['queue'] = queue
            else:
                try:
                    route['queue'] = self.queues[queue]
                except KeyError:
                    raise QueueNotFound(
                        f'Queue {queue!r} missing from task_queues')
        return route

    def lookup_route(self, name,
                     args=None, kwargs=None, options=None, task_type=None):
        query = self.query_router
        for router in self.routes:
            route = query(router, name, args, kwargs, options, task_type)
            if route is not None:
                return route

    def query_router(self, router, task, args, kwargs, options, task_type):
        router = maybe_evaluate(router)
        if hasattr(router, 'route_for_task'):
            # pre 4.0 router class
            return router.route_for_task(task, args, kwargs)
        return router(task, args, kwargs, options, task=task_type)


def expand_router_string(router):
    router = symbol_by_name(router)
    if hasattr(router, 'route_for_task'):
        # need to instantiate pre 4.0 router classes
        router = router()
    return router


def prepare(routes):
    """Expand the :setting:`task_routes` setting."""

    def expand_route(route):
        if isinstance(route, (Mapping, list, tuple)):
            return MapRoute(route)
        if isinstance(route, str):
            return mlazy(expand_router_string, route)
        return route

    if routes is None:
        return ()
    if not isinstance(routes, (list, tuple)):
        routes = (routes,)
    return [expand_route(route) for route in routes]



import warnings
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from . import registry
from .event import Event
from .exceptions import InvalidDefinition
from .graph import iterate_states_and_transitions
from .graph import visit_connected_states
from .i18n import _
from .state import State
from .states import States
from .transition import Transition
from .transition_list import TransitionList


class StateMachineMetaclass(type):
    "Metaclass for constructing StateMachine classes"

    def __init__(
        cls,
        name: str,
        bases: Tuple[type],
        attrs: Dict[str, Any],
        strict_states: bool = False,
    ) -> None:
        super().__init__(name, bases, attrs)
        registry.register(cls)
        cls.name = cls.__name__
        cls.states: States = States()
        cls.states_map: Dict[Any, State] = {}
        """Map of ``state.value`` to the corresponding :ref:`state`."""

        cls._abstract = True
        cls._strict_states = strict_states
        cls._events: Dict[Event, None] = {}  # used Dict to preserve order and avoid duplicates
        cls._protected_attrs: set = set()
        cls._events_to_update: Dict[Event, Event | None] = {}

        cls.add_inherited(bases)
        cls.add_from_attributes(attrs)
        cls._update_event_references()

        try:
            cls.initial_state: State = next(s for s in cls.states if s.initial)
        except StopIteration:
            cls.initial_state = None  # Abstract SM still don't have states

        cls.final_states: List[State] = [state for state in cls.states if state.final]

        cls._check()
        cls._setup()

    if TYPE_CHECKING:
        """Makes mypy happy with dynamic created attributes"""

        def __getattr__(self, attribute: str) -> Any: ...

    def _check(cls):
        has_states = bool(cls.states)
        has_events = bool(cls._events)

        cls._abstract = not has_states and not has_events

        # do not validate the base abstract classes
        if cls._abstract:
            return

        if not has_states:
            raise InvalidDefinition(_("There are no states."))

        if not has_events:
            raise InvalidDefinition(_("There are no events."))

        cls._check_initial_state()
        cls._check_final_states()
        cls._check_disconnected_state()
        cls._check_trap_states()
        cls._check_reachable_final_states()

    def _check_initial_state(cls):
        initials = [s for s in cls.states if s.initial]
        if len(initials) != 1:
            raise InvalidDefinition(
                _(
                    "There should be one and only one initial state. "
                    "You currently have these: {!r}"
                ).format([s.id for s in initials])
            )

    def _check_final_states(cls):
        final_state_with_invalid_transitions = [
            state for state in cls.final_states if state.transitions
        ]

        if final_state_with_invalid_transitions:
            raise InvalidDefinition(
                _("Cannot declare transitions from final state. Invalid state(s): {}").format(
                    [s.id for s in final_state_with_invalid_transitions]
                )
            )

    def _check_trap_states(cls):
        trap_states = [s for s in cls.states if not s.final and not s.transitions]
        if trap_states:
            message = _(
                "All non-final states should have at least one outgoing transition. "
                "These states have no outgoing transition: {!r}"
            ).format([s.id for s in trap_states])
            if cls._strict_states:
                raise InvalidDefinition(message)
            else:
                warnings.warn(message, UserWarning, stacklevel=4)

    def _check_reachable_final_states(cls):
        if not any(s.final for s in cls.states):
            return  # No need to check final reachability
        disconnected_states = cls._states_without_path_to_final_states()
        if disconnected_states:
            message = _(
                "All non-final states should have at least one path to a final state. "
                "These states have no path to a final state: {!r}"
            ).format([s.id for s in disconnected_states])
            if cls._strict_states:
                raise InvalidDefinition(message)
            else:
                warnings.warn(message, UserWarning, stacklevel=1)

    def _states_without_path_to_final_states(cls):
        return [
            state
            for state in cls.states
            if not state.final and not any(s.final for s in visit_connected_states(state))
        ]

    def _disconnected_states(cls, starting_state):
        visitable_states = set(visit_connected_states(starting_state))
        return set(cls.states) - visitable_states

    def _check_disconnected_state(cls):
        disconnected_states = cls._disconnected_states(cls.initial_state)
        if disconnected_states:
            raise InvalidDefinition(
                _(
                    "There are unreachable states. "
                    "The statemachine graph should have a single component. "
                    "Disconnected states: {}"
                ).format([s.id for s in disconnected_states])
            )

    def _setup(cls):
        for visited in iterate_states_and_transitions(cls.states):
            visited._setup()

        cls._protected_attrs = {
            "_abstract",
            "model",
            "state_field",
            "start_value",
            "initial_state",
            "final_states",
            "states",
            "_events",
            "states_map",
            "send",
        } | {s.id for s in cls.states}

    def add_inherited(cls, bases):
        for base in bases:
            for state in getattr(base, "states", []):
                cls.add_state(state.id, state)

            events = getattr(base, "_events", {})
            for event in events:
                cls.add_event(event=Event(id=event.id, name=event.name))

    def add_from_attributes(cls, attrs):  # noqa: C901
        for key, value in attrs.items():
            if isinstance(value, States):
                cls._add_states_from_dict(value)
            if isinstance(value, State):
                cls.add_state(key, value)
            elif isinstance(value, (Transition, TransitionList)):
                cls.add_event(event=Event(transitions=value, id=key, name=key))
            elif isinstance(value, (Event,)):
                cls.add_event(
                    event=Event(
                        transitions=value._transitions,
                        id=key,
                        name=value.name,
                    ),
                    old_event=value,
                )
            elif getattr(value, "attr_name", None):
                cls._add_unbounded_callback(key, value)

    def _add_states_from_dict(cls, states):
        for state_id, state in states.items():
            cls.add_state(state_id, state)

    def _add_unbounded_callback(cls, attr_name, func):
        # if func is an event, the `attr_name` will be replaced by an event trigger,
        # so we'll also give the ``func`` a new unique name to be used by the callback
        # machinery that is stored at ``func.attr_name``
        setattr(cls, func.attr_name, func)
        if func.is_event:
            cls.add_event(event=Event(func._transitions, id=attr_name, name=attr_name))

    def add_state(cls, id, state: State):
        state._set_id(id)
        cls.states.append(state)
        cls.states_map[state.value] = state
        if not hasattr(cls, id):
            setattr(cls, id, state)

        # also register all events associated directly with transitions
        for event in state.transitions.unique_events:
            cls.add_event(event)

    def add_event(
        cls,
        event: Event,
        old_event: "Event | None" = None,
    ):
        if not event._has_real_id:
            if event not in cls._events_to_update:
                cls._events_to_update[event] = None
            return

        transitions = event._transitions
        if transitions is not None:
            transitions._on_event_defined(event=event, states=list(cls.states))

        if event not in cls._events:
            cls._events[event] = None
            setattr(cls, event.id, event)

        if old_event is not None:
            cls._events_to_update[old_event] = event

        return cls._events[event]

    def _update_event_references(cls):
        for old_event, new_event in cls._events_to_update.items():
            for state in cls.states:
                for transition in state.transitions:
                    if transition._events.match(old_event):
                        if new_event is None:
                            raise InvalidDefinition(
                                _("An event in the '{}' has no id.").format(transition)
                            )
                        transition.events._replace(old_event, new_event)

        cls._events_to_update = {}

    @property
    def events(self):
        return list(self._events)



import os


def load_ctime_functions():
    if os.name == "nt":
        import win32_setctime

        def get_ctime_windows(filepath):
            return os.stat(filepath).st_ctime

        def set_ctime_windows(filepath, timestamp):
            if not win32_setctime.SUPPORTED:
                return

            try:
                win32_setctime.setctime(filepath, timestamp)
            except (OSError, ValueError):
                pass

        return get_ctime_windows, set_ctime_windows

    if hasattr(os.stat_result, "st_birthtime"):

        def get_ctime_macos(filepath):
            return os.stat(filepath).st_birthtime

        def set_ctime_macos(filepath, timestamp):
            pass

        return get_ctime_macos, set_ctime_macos

    if hasattr(os, "getxattr") and hasattr(os, "setxattr"):

        def get_ctime_linux(filepath):
            try:
                return float(os.getxattr(filepath, b"user.loguru_crtime"))
            except OSError:
                return os.stat(filepath).st_mtime

        def set_ctime_linux(filepath, timestamp):
            try:
                os.setxattr(filepath, b"user.loguru_crtime", str(timestamp).encode("ascii"))
            except OSError:
                pass

        return get_ctime_linux, set_ctime_linux

    def get_ctime_fallback(filepath):
        return os.stat(filepath).st_mtime

    def set_ctime_fallback(filepath, timestamp):
        pass

    return get_ctime_fallback, set_ctime_fallback


get_ctime, set_ctime = load_ctime_functions()



"""
    ulid/codec
    ~~~~~~~~~~

    Defines encoding/decoding functions for ULID data representations.
"""
import datetime
import typing

from . import base32, hints, ulid

#: Type hint that defines multiple primitive types that can represent
#: a Unix timestamp in seconds.
TimestampPrimitive = typing.Union[hints.Primitive,  # pylint: disable=invalid-name
                                  datetime.datetime, ulid.Timestamp, ulid.ULID]


#: Type hint that defines multiple primitive types that can represent randomness.
RandomnessPrimitive = typing.Union[hints.Primitive, ulid.Randomness, ulid.ULID]  # pylint: disable=invalid-name


def decode_timestamp(timestamp: TimestampPrimitive) -> ulid.Timestamp:
    """
    Create a new :class:`~ulid.ulid.ULID` instance using a timestamp value of a supported type.

    The following types are supported for timestamp values:

    * :class:`~datetime.datetime`
    * :class:`~int`
    * :class:`~float`
    * :class:`~str`
    * :class:`~memoryview`
    * :class:`~ulid.ulid.Timestamp`
    * :class:`~ulid.ulid.ULID`
    * :class:`~bytes`
    * :class:`~bytearray`

    :param timestamp: Unix timestamp in seconds
    :type timestamp: See docstring for types
    :return: ULID using given timestamp and new randomness
    :rtype: :class:`~ulid.ulid.ULID`
    :raises ValueError: when the value is an unsupported type
    :raises ValueError: when the value is a string and cannot be Base32 decoded
    :raises ValueError: when the value is or was converted to something 48 bits
    """
    if isinstance(timestamp, datetime.datetime):
        timestamp = timestamp.timestamp()
    if isinstance(timestamp, (int, float)):
        timestamp = int(timestamp * 1000.0).to_bytes(6, byteorder='big')
    elif isinstance(timestamp, str):
        timestamp = base32.decode_timestamp(timestamp)
    elif isinstance(timestamp, memoryview):
        timestamp = timestamp.tobytes()
    elif isinstance(timestamp, ulid.Timestamp):
        timestamp = timestamp.bytes
    elif isinstance(timestamp, ulid.ULID):
        timestamp = timestamp.timestamp().bytes

    if not isinstance(timestamp, (bytes, bytearray)):
        raise ValueError('Expected datetime, int, float, str, memoryview, Timestamp, ULID, '
                         'bytes, or bytearray; got {}'.format(type(timestamp).__name__))

    length = len(timestamp)
    if length != 6:
        raise ValueError('Expects timestamp to be 48 bits; got {} bytes'.format(length))

    return ulid.Timestamp(timestamp)


def decode_randomness(randomness: RandomnessPrimitive) -> ulid.Randomness:
    """
    Create a new :class:`~ulid.ulid.Randomness` instance using the given randomness value of a supported type.

    The following types are supported for randomness values:

    * :class:`~int`
    * :class:`~float`
    * :class:`~str`
    * :class:`~memoryview`
    * :class:`~ulid.ulid.Randomness`
    * :class:`~ulid.ulid.ULID`
    * :class:`~bytes`
    * :class:`~bytearray`

    :param randomness: Random bytes
    :type randomness: See docstring for types
    :return: ULID using new timestamp and given randomness
    :rtype: :class:`~ulid.ulid.ULID`
    :raises ValueError: when the value is an unsupported type
    :raises ValueError: when the value is a string and cannot be Base32 decoded
    :raises ValueError: when the value is or was converted to something 80 bits
    """
    if isinstance(randomness, (int, float)):
        randomness = int(randomness).to_bytes(10, byteorder='big')
    elif isinstance(randomness, str):
        randomness = base32.decode_randomness(randomness)
    elif isinstance(randomness, memoryview):
        randomness = randomness.tobytes()
    elif isinstance(randomness, ulid.Randomness):
        randomness = randomness.bytes
    elif isinstance(randomness, ulid.ULID):
        randomness = randomness.randomness().bytes

    if not isinstance(randomness, (bytes, bytearray)):
        raise ValueError('Expected int, float, str, memoryview, Randomness, ULID, '
                         'bytes, or bytearray; got {}'.format(type(randomness).__name__))

    length = len(randomness)
    if length != 10:
        raise ValueError('Expects randomness to be 80 bits; got {} bytes'.format(length))

    return ulid.Randomness(randomness)



"""The ``RPC`` result backend for AMQP brokers.

RPC-style result backend, using reply-to and one queue per client.
"""
import time

import kombu
from kombu.common import maybe_declare
from kombu.utils.compat import register_after_fork
from kombu.utils.objects import cached_property

from celery import states
from celery._state import current_task, task_join_will_block

from . import base
from .asynchronous import AsyncBackendMixin, BaseResultConsumer

__all__ = ('BacklogLimitExceeded', 'RPCBackend')

E_NO_CHORD_SUPPORT = """
The "rpc" result backend does not support chords!

Note that a group chained with a task is also upgraded to be a chord,
as this pattern requires synchronization.

Result backends that supports chords: Redis, Database, Memcached, and more.
"""


class BacklogLimitExceeded(Exception):
    """Too much state history to fast-forward."""


def _on_after_fork_cleanup_backend(backend):
    backend._after_fork()


class ResultConsumer(BaseResultConsumer):
    Consumer = kombu.Consumer

    _connection = None
    _consumer = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._create_binding = self.backend._create_binding

    def start(self, initial_task_id, no_ack=True, **kwargs):
        self._connection = self.app.connection()
        initial_queue = self._create_binding(initial_task_id)
        self._consumer = self.Consumer(
            self._connection.default_channel, [initial_queue],
            callbacks=[self.on_state_change], no_ack=no_ack,
            accept=self.accept)
        self._consumer.consume()

    def drain_events(self, timeout=None):
        if self._connection:
            return self._connection.drain_events(timeout=timeout)
        elif timeout:
            time.sleep(timeout)

    def stop(self):
        try:
            self._consumer.cancel()
        finally:
            self._connection.close()

    def on_after_fork(self):
        self._consumer = None
        if self._connection is not None:
            self._connection.collect()
            self._connection = None

    def consume_from(self, task_id):
        if self._consumer is None:
            return self.start(task_id)
        queue = self._create_binding(task_id)
        if not self._consumer.consuming_from(queue):
            self._consumer.add_queue(queue)
            self._consumer.consume()

    def cancel_for(self, task_id):
        if self._consumer:
            self._consumer.cancel_by_queue(self._create_binding(task_id).name)


class RPCBackend(base.Backend, AsyncBackendMixin):
    """Base class for the RPC result backend."""

    Exchange = kombu.Exchange
    Producer = kombu.Producer
    ResultConsumer = ResultConsumer

    #: Exception raised when there are too many messages for a task id.
    BacklogLimitExceeded = BacklogLimitExceeded

    persistent = False
    supports_autoexpire = True
    supports_native_join = True

    retry_policy = {
        'max_retries': 20,
        'interval_start': 0,
        'interval_step': 1,
        'interval_max': 1,
    }

    class Consumer(kombu.Consumer):
        """Consumer that requires manual declaration of queues."""

        auto_declare = False

    class Queue(kombu.Queue):
        """Queue that never caches declaration."""

        can_cache_declaration = False

    def __init__(self, app, connection=None, exchange=None, exchange_type=None,
                 persistent=None, serializer=None, auto_delete=True, **kwargs):
        super().__init__(app, **kwargs)
        conf = self.app.conf
        self._connection = connection
        self._out_of_band = {}
        self.persistent = self.prepare_persistent(persistent)
        self.delivery_mode = 2 if self.persistent else 1
        exchange = exchange or conf.result_exchange
        exchange_type = exchange_type or conf.result_exchange_type
        self.exchange = self._create_exchange(
            exchange, exchange_type, self.delivery_mode,
        )
        self.serializer = serializer or conf.result_serializer
        self.auto_delete = auto_delete
        self.result_consumer = self.ResultConsumer(
            self, self.app, self.accept,
            self._pending_results, self._pending_messages,
        )
        if register_after_fork is not None:
            register_after_fork(self, _on_after_fork_cleanup_backend)

    def _after_fork(self):
        # clear state for child processes.
        self._pending_results.clear()
        self.result_consumer._after_fork()

    def _create_exchange(self, name, type='direct', delivery_mode=2):
        # uses direct to queue routing (anon exchange).
        return self.Exchange(None)

    def _create_binding(self, task_id):
        """Create new binding for task with id."""
        # RPC backend caches the binding, as one queue is used for all tasks.
        return self.binding

    def ensure_chords_allowed(self):
        raise NotImplementedError(E_NO_CHORD_SUPPORT.strip())

    def on_task_call(self, producer, task_id):
        # Called every time a task is sent when using this backend.
        # We declare the queue we receive replies on in advance of sending
        # the message, but we skip this if running in the prefork pool
        # (task_join_will_block), as we know the queue is already declared.
        if not task_join_will_block():
            maybe_declare(self.binding(producer.channel), retry=True)

    def destination_for(self, task_id, request):
        """Get the destination for result by task id.

        Returns:
            Tuple[str, str]: tuple of ``(reply_to, correlation_id)``.
        """
        # Backends didn't always receive the `request`, so we must still
        # support old code that relies on current_task.
        try:
            request = request or current_task.request
        except AttributeError:
            raise RuntimeError(
                f'RPC backend missing task request for {task_id!r}')
        return request.reply_to, request.correlation_id or task_id

    def on_reply_declare(self, task_id):
        # Return value here is used as the `declare=` argument
        # for Producer.publish.
        # By default we don't have to declare anything when sending a result.
        pass

    def on_result_fulfilled(self, result):
        # This usually cancels the queue after the result is received,
        # but we don't have to cancel since we have one queue per process.
        pass

    def as_uri(self, include_password=True):
        return 'rpc://'

    def store_result(self, task_id, result, state,
                     traceback=None, request=None, **kwargs):
        """Send task return value and state."""
        routing_key, correlation_id = self.destination_for(task_id, request)
        if not routing_key:
            return
        with self.app.amqp.producer_pool.acquire(block=True) as producer:
            producer.publish(
                self._to_result(task_id, state, result, traceback, request),
                exchange=self.exchange,
                routing_key=routing_key,
                correlation_id=correlation_id,
                serializer=self.serializer,
                retry=True, retry_policy=self.retry_policy,
                declare=self.on_reply_declare(task_id),
                delivery_mode=self.delivery_mode,
            )
        return result

    def _to_result(self, task_id, state, result, traceback, request):
        return {
            'task_id': task_id,
            'status': state,
            'result': self.encode_result(result, state),
            'traceback': traceback,
            'children': self.current_task_children(request),
        }

    def on_out_of_band_result(self, task_id, message):
        # Callback called when a reply for a task is received,
        # but we have no idea what to do with it.
        # Since the result is not pending, we put it in a separate
        # buffer: probably it will become pending later.
        if self.result_consumer:
            self.result_consumer.on_out_of_band_result(message)
        self._out_of_band[task_id] = message

    def get_task_meta(self, task_id, backlog_limit=1000):
        buffered = self._out_of_band.pop(task_id, None)
        if buffered:
            return self._set_cache_by_message(task_id, buffered)

        # Polling and using basic_get
        latest_by_id = {}
        prev = None
        for acc in self._slurp_from_queue(task_id, self.accept, backlog_limit):
            tid = self._get_message_task_id(acc)
            prev, latest_by_id[tid] = latest_by_id.get(tid), acc
            if prev:
                # backends aren't expected to keep history,
                # so we delete everything except the most recent state.
                prev.ack()
                prev = None

        latest = latest_by_id.pop(task_id, None)
        for tid, msg in latest_by_id.items():
            self.on_out_of_band_result(tid, msg)

        if latest:
            latest.requeue()
            return self._set_cache_by_message(task_id, latest)
        else:
            # no new state, use previous
            try:
                return self._cache[task_id]
            except KeyError:
                # result probably pending.
                return {'status': states.PENDING, 'result': None}
    poll = get_task_meta  # XXX compat

    def _set_cache_by_message(self, task_id, message):
        payload = self._cache[task_id] = self.meta_from_decoded(
            message.payload)
        return payload

    def _slurp_from_queue(self, task_id, accept,
                          limit=1000, no_ack=False):
        with self.app.pool.acquire_channel(block=True) as (_, channel):
            binding = self._create_binding(task_id)(channel)
            binding.declare()

            for _ in range(limit):
                msg = binding.get(accept=accept, no_ack=no_ack)
                if not msg:
                    break
                yield msg
            else:
                raise self.BacklogLimitExceeded(task_id)

    def _get_message_task_id(self, message):
        try:
            # try property first so we don't have to deserialize
            # the payload.
            return message.properties['correlation_id']
        except (AttributeError, KeyError):
            # message sent by old Celery version, need to deserialize.
            return message.payload['task_id']

    def revive(self, channel):
        pass

    def reload_task_result(self, task_id):
        raise NotImplementedError(
            'reload_task_result is not supported by this backend.')

    def reload_group_result(self, task_id):
        """Reload group result, even if it has been previously fetched."""
        raise NotImplementedError(
            'reload_group_result is not supported by this backend.')

    def save_group(self, group_id, result):
        raise NotImplementedError(
            'save_group is not supported by this backend.')

    def restore_group(self, group_id, cache=True):
        raise NotImplementedError(
            'restore_group is not supported by this backend.')

    def delete_group(self, group_id):
        raise NotImplementedError(
            'delete_group is not supported by this backend.')

    def __reduce__(self, args=(), kwargs=None):
        kwargs = {} if not kwargs else kwargs
        return super().__reduce__(args, dict(
            kwargs,
            connection=self._connection,
            exchange=self.exchange.name,
            exchange_type=self.exchange.type,
            persistent=self.persistent,
            serializer=self.serializer,
            auto_delete=self.auto_delete,
            expires=self.expires,
        ))

    @property
    def binding(self):
        return self.Queue(
            self.oid, self.exchange, self.oid,
            durable=False,
            auto_delete=True,
            expires=self.expires,
        )

    @cached_property
    def oid(self):
        # cached here is the app thread OID: name of queue we receive results on.
        return self.app.thread_oid



from dependency_injector import containers, providers
from redis import Redis

from genie_flow.session_lock import SessionLockManager


class GenieFlowPersistenceContainer(containers.DeclarativeContainer):

    config = providers.Configuration()

    redis_object_store = providers.Singleton(
        Redis,
        host=config.object_store.host,
        port=config.object_store.port,
        db=config.object_store.db,
        password=config.object_store.password,
    )

    redis_lock_store = providers.Singleton(
        Redis,
        host=config.lock_store.host,
        port=config.lock_store.port,
        db=config.lock_store.db,
        password=config.lock_store.password,
    )

    redis_progress_store = providers.Singleton(
        Redis,
        host=config.progress_store.host,
        port=config.progress_store.port,
        db=config.progress_store.db,
        password=config.progress_store.password,
    )

    session_lock_manager = providers.Singleton(
        SessionLockManager,
        redis_object_store=redis_object_store,
        redis_lock_store=redis_lock_store,
        redis_progress_store=redis_progress_store,
        compression=config.object_store.object_compression or True,
        application_prefix=config.application_prefix or 'genie-flow',
        object_expiration_seconds=config.object_store.expiration_seconds or 120,
        lock_expiration_seconds=config.lock_store.expiration_seconds or 120,
        progress_expiration_seconds=config.progress_store.expiration_seconds or 120,
    )



import json
from typing import Any, Optional

import jmespath
from celery import Celery, Task
from celery.app.task import Context
from celery.canvas import chord, group, Signature
from celery.result import AsyncResult
from loguru import logger
from statemachine import State

from genie_flow.celery.compiler import TaskCompiler
from genie_flow.celery.progress import ProgressLoggingTask
from genie_flow.celery.transition import TransitionManager
from genie_flow.environment import GenieEnvironment
from genie_flow.genie import GenieModel, GenieStateMachine
from genie_flow.model.template import CompositeContentType
from genie_flow.mongo import store_session, store_user
from genie_flow.session_lock import SessionLockManager
from genie_flow.utils import get_fully_qualified_name_from_class


def parse_if_json(s: str) -> Any:
    if not isinstance(s, str):
        return s

    try:
        return json.loads(s)
    except (json.JSONDecodeError, UnicodeDecodeError):
        return s


class CeleryManager:
    """
    The `CeleryManager` instance deals with compiling and enqueuing Celery tasks.
    """

    def __init__(
        self,
        celery: Celery,
        session_lock_manager: SessionLockManager,
        genie_environment: GenieEnvironment,
        update_mongo_period: float,
    ):
        self.celery_app = celery
        self.session_lock_manager = session_lock_manager
        self.genie_environment = genie_environment
        self.update_mongo_period = update_mongo_period

        self._add_error_handler()
        self._add_trigger_ai_event_task()
        self._add_invoke_task()
        self._add_wrap_index()
        self._add_recompile()
        self._add_map_task()
        self._add_combine_group_to_dict()
        self._add_combine_group_to_list()
        self._add_chained_template()
        self._add_update_mongo_task()
        self._add_periodic_tasks()

    def _retrieve_render_data(
            self,
            drag_net: Optional[dict],
            session_id: str,
            model_fqn: str,
    ) -> dict:
        """
        Retrieve the render data for a given session id and model. A task may have run before
        this and created a drag net dictionary of attributes that need to be added to the
        `render_data`.

        :param drag_net: an optional dict of values that need to be merged into the `render_data`
        :param session_id: the session id
        :param model_fqn: the model fully-qualified name
        :return: a dict with render data
        """
        logger.debug(
            "Retrieving render data for session {session_id}",
            session_id=session_id,
        )
        model = self.session_lock_manager.get_model(session_id, model_fqn)
        render_data = model.render_data

        if drag_net is not None:
            logger.debug(
                "Merging drag net values {drag_net} with render data",
                drag_net=drag_net,
            )
            render_data.update(drag_net)
        return render_data

    def _process_model_event(
            self,
            event_argument: str,
            model: GenieModel,
            session_id: str,
            event_name: str,
    ):
        state_machine = model.get_state_machine_class()(model)
        state_machine.add_listener(TransitionManager(self))

        logger.debug(f"sending {event_name} to model for session {session_id}")
        state_machine.send(event_name, event_argument)
        logger.debug(f"actor input is now {model.actor_input}")

    def _add_error_handler(self):

        @self.celery_app.task(name="genie_flow.error_handler")
        def error_handler(
                request: Context,
                exc,
                traceback,
                cls_fqn: str,
                session_id: str,
                invocation_id: str,
                event_name: str,
        ):
            """
            Process a backend error. The error is captured and the exception added to the model's
            task_error property. The final event is (still) being sent to the state machine. But the
            actor's input is an empty string.
            """
            logger.error(
                "Task {request.id}, for session {session_id}, invocation {invocation_id} "
                "raised an error: {exc}",
                request=request,
                session_id=session_id,
                invocation_id=invocation_id,
                exc=exc,
            )
            logger.exception(traceback)

            with self.session_lock_manager.get_locked_model(session_id, cls_fqn) as model:
                self._process_model_event(
                    event_argument="",
                    model=model,
                    session_id=session_id,
                    event_name=event_name,
                )

                if model.task_error is None:
                    model.task_error = ""
                model.task_error += json.dumps(
                    dict(
                        session_id=session_id,
                        invocation_id=invocation_id,
                        task_id=request.id,
                        task_name=request.id,
                        exception=str(exc),
                    )
                )

        return error_handler

    def _add_trigger_ai_event_task(self):

        @self.celery_app.task(
            bind=True,
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name='genie_flow.trigger_ai_event'
        )
        def trigger_ai_event(
                task_instance,
                response: str,
                event_name: str,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ):
            """
            This Celery Task is executed at the end of a Celery DAG and all the relevant
            Invokers have run. It takes the output of the previous task, pulls up the model
            form the store, creates the state machine for it and sends that state machine
            the event that was given.

            :param task_instance: Celery Task instance - a reference to this task itself (bound)
            :param response: The response from the previous task
            :param event_name: The name of the event that needs to be sent to the state machine
            :param session_id: The session id for which this task is executed
            :param model_fqn: The fully qualified name of the class of the model
            """
            with self.session_lock_manager.get_locked_model(session_id, model_fqn) as model:
                self.session_lock_manager.progress_tombstone(session_id, invocation_id)

                state_machine = model.get_state_machine_class()(model)
                state_machine.add_listener(TransitionManager(self))

                logger.debug(
                    "sending {event_name} to state machine for session {session_id}",
                    event_name=event_name,
                    session_id=session_id,
                )
                state_machine.send(event_name, response)

                if model.actor_input is None:
                    logger.debug("actor input is None")
                else:
                    logger.debug(
                        "actor input is now '{actor_input}'",
                        actor_input=(
                            model.actor_input
                            if len(model.actor_input) < 50
                            else model.actor_input[:50] + "..."
                        ),
                    )

        return trigger_ai_event

    def _add_invoke_task(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.invoke_task",
        )
        def invoke_ai_event(
                drag_net: Optional[dict],
                template_name: str,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ) -> str:
            """
            This Celery Task executes the actual Invocation. It is given the data that should be
            used to render the template. It then invokes the template.

            :param drag_net: potential dict of values that need to be merged into the `render_data`
            :param template_name: The name of the template that should be used to render
            :param session_id: The session id for which this task is executed
            :param model_fqn: The fully qualified name of the model
            :param invocation_id: the id of the invocation that is being executed
            :returns: the result of the invocation
            """
            render_data = self._retrieve_render_data(drag_net, session_id, model_fqn)
            return self.genie_environment.invoke_template(template_name, render_data)

        return invoke_ai_event

    def _add_wrap_index(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.wrap_index",
        )
        def wrap_index(
                map_index: int,
                task_signature: Signature | dict,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ) -> tuple[int, Any]:
            """
            This is a helper function that wraps the invocation of a task with the index of
            the position in a list. This is used to ensure that the order of the results can
            be recompiled correctly.

            :param map_index: the position in the list of the invocation result
            :param task_signature: the signature of the task to be invoked, or a dict of the same
            :param session_id: The session id for which this task is executed
            :param model_fqn: The fully qualified name of the model
            :param invocation_id: the id of the invocation that is being executed
            :return: a tuple containing the index and the result of the invocation.
            """
            if isinstance(task_signature, dict):
                task_signature = Signature.from_dict(task_signature)
            return map_index, parse_if_json(task_signature())

        return wrap_index

    def _add_recompile(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.recompile",
        )
        def recompile(
                results: list,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ):
            """
            This is a helper function that re-orders the results of a task that has been
            jumbled. This function expects a list of tuples, where the first element is the
            original position in a list. The second element is the result of the invocation.
            :param results: a list of tuples
            :param session_id: the session id for which this task is executed
            :param model_fqn: the model fully qualified name for which this task is executed
            :param invocation_id: the invocation id for which this task is executed
            :return: a list of results in the order they were invoked.
            """
            if len(results) > 1:
                if all(
                        results[i][0] == results[i-1][0] + 1
                        for i in range(1, len(results))
                ):
                    logger.warning(
                        "results are in order, reordering not strictly required "
                        "for session {session_id} invocation {invocation_id}",
                        session_id=session_id,
                        invocation_id=invocation_id,
                    )

            results.sort(key=lambda x: x[0])
            return json.dumps([r[1] for r in results])

        return recompile

    def _add_map_task(self):

        @self.celery_app.task(
            bind=True,
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.map_task",
        )
        def map_task(
                task_instance: Task,
                drag_net: Optional[dict],
                list_attribute: str,
                map_index_field: str,
                map_value_field: str,
                template_name: str,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ):
            """
            This task maps a template onto the different values in a list of model parameters.
            Each of the invocations will be created as a separate Celery task. A final task
            will be run, converting the output into a JSON list of results.

            This mapping will be done at run-time, so all the values in the model's list
            attribute will generate a separate invocation of the template.

            When the template is invoked, it will be rendered with the complete render_data
            object, with an addition of two attributes: the attribute identifying the index
            of the value it is rendered for, and an attribute containing the value itself.

            The names of these attributes are given by `map_index_field` and `map_value_field`
            respectively.

            At this time, only a simple rendered template can be used - no list, dict or
            otherwise.

            :param task_instance: a reference to the map task itself
            :param drag_net: potential dict of values that need to be merged into the `render_data`
            :param list_attribute: the JMES Path into the attribute to map
            :param map_index_field: the name of the attribute carrying the index
            :param map_value_field: the name of the attribute carrying the value
            :param template_name: the name of the template that should be used to render
            :param session_id: the session id for which this task is executed
            :param model_fqn: the fully qualified name of the model
            """
            render_data = self._retrieve_render_data(drag_net, session_id, model_fqn)
            list_values = jmespath.search(list_attribute, render_data)
            if not isinstance(list_values, list):
                logger.warning(
                    "path to attribute returns type {path_type} and not a list",
                    path_type=type(list_values),
                )
                list_values = [list_values]

            index_wrapper_task = self.celery_app.tasks["genie_flow.wrap_index"]
            invoke_task = self.celery_app.tasks["genie_flow.invoke_task"]

            # We wrap the actual invocation with the index wrapper. This is to work around
            # an apparent bug in Celery where replacing a task with a group jumbles up the
            # order of the results. See the discussion on the Celery home here:
            # https://github.com/celery/celery/discussions/9731
            # The index wrapper task simply returns the index and the invocation result.
            # We then use the recompile task to re-order the results.
            mapped_tasks = [
                index_wrapper_task.s(
                    map_index,
                    invoke_task.s(
                        {
                            map_index_field: map_index,
                            map_value_field: map_value,
                        },
                        template_name,
                        session_id,
                        model_fqn,
                        invocation_id,
                    ),
                    session_id,
                    model_fqn,
                    invocation_id,
                )
                for map_index, map_value in enumerate(list_values)
            ]
            recompile_task = self.celery_app.tasks["genie_flow.recompile"]

            # increase the number of tasks To Do by the number of values mapped,
            # plus one for the combine task, minus one because we are replacing
            # this MapTaskTemplate task that was already counted
            self.session_lock_manager.progress_update_todo(
                session_id,
                invocation_id,
                len(list_values),
            )
            return task_instance.replace(
                chord(
                    group(*mapped_tasks),
                    recompile_task.s(session_id, model_fqn, invocation_id),
                )
            )

        return map_task

    def _add_combine_group_to_dict(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.combine_group_to_dict",
        )
        def combine_group_to_dict(
                results: list[CompositeContentType],
                keys: list[str],
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ) -> CompositeContentType:
            parsed_results = [parse_if_json(s) for s in results]
            return json.dumps(dict(zip(keys, parsed_results)))

        return combine_group_to_dict

    def _add_combine_group_to_list(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.combine_group_to_list",
        )
        def combine_chain_to_list(
                results: list[CompositeContentType],
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ):
            parsed_results = [parse_if_json(s) for s in results]
            return json.dumps(parsed_results)

        return combine_chain_to_list

    def _add_chained_template(self):

        @self.celery_app.task(
            base=ProgressLoggingTask,
            session_lock_manager=self.session_lock_manager,
            name="genie_flow.chained_template",
        )
        def chained_template(
                result_of_previous_call: CompositeContentType,
                session_id: str,
                model_fqn: str,
                invocation_id: str,
        ) -> CompositeContentType:

            parsed_previous_result = None
            try:
                parsed_previous_result = json.loads(result_of_previous_call)
            except json.decoder.JSONDecodeError:
                pass

            return dict(
                previous_result=result_of_previous_call,
                parsed_previous_result=parsed_previous_result,
            )

        return chained_template

    def enqueue_task(
            self,
            state_machine: GenieStateMachine,
            model: GenieModel,
            target_state: State,
    ):
        """
        Create a new Celery DAG and place it on the Celery queue.

        The DAG is compiled using the `TaskCompiler`, the error handler gets assigned,
        the DAG is enqueued and a new `GenieTaskProgress` object is persisted.

        This is also the point in time where the `render_data` is created (by using the
        `render_data` property of the machine) and therefore frozen. That then becomes
        the `render_data` that is used inside the DAG.

        :param state_machine: the active state machine to use
        :param model: the data model
        :param target_state: the state we will transition into
        """
        model_fqn = get_fully_qualified_name_from_class(model)
        event_to_send_after = target_state.transitions.unique_events[0]
        task_compiler = TaskCompiler(
            self.celery_app,
            state_machine.get_template_for_state(target_state),
            model.session_id,
            model_fqn,
            target_state.id,
            event_to_send_after,
        )
        task_compiler.task.on_error(
            task_compiler.error_handler.s(
                model_fqn,
                model.session_id,
                task_compiler.invocation_id,
                event_to_send_after,
            )
        )

        # enqueuing the compiled task with an empty drag_net dictionary
        task = task_compiler.task.apply_async((None,))

        self.session_lock_manager.progress_start(
            session_id=model.session_id,
            invocation_id=task_compiler.invocation_id,
            nr_tasks_todo=task_compiler.nr_tasks,
        )

    def get_task_result(self, task_id) -> AsyncResult:
        return AsyncResult(task_id, app=self.celery_app)

    def _add_update_mongo_task(self):

        @self.celery_app.task(name="genie_flow.scheduler.update_mongo")
        def update_mongo():
            logger.debug("update mongo running")
            updated_sessions = self.session_lock_manager.redis_object_store.smembers(self.session_lock_manager.update_set_key)
            for session_item in updated_sessions:
                fqn, session_id = session_item.decode().rsplit(':', 1)
                with self.session_lock_manager.get_locked_model(
                        session_id=session_id,
                        model_class=fqn
                ) as model:
                    store_session(model)
                    store_user(model.secondary_storage["user_info"], session_id)
                    self.session_lock_manager.redis_object_store.srem(self.session_lock_manager.update_set_key, session_id)
        return update_mongo

    def _add_periodic_tasks(self):
        self.celery_app.conf.beat_schedule = {
            'add-every-30-seconds': {
                'task': 'genie_flow.scheduler.update_mongo',
                'schedule': self.update_mongo_period
            },
        }


from typing import TYPE_CHECKING
from typing import Iterable
from typing import List

from .callbacks import CallbackGroup
from .transition import Transition
from .transition_mixin import AddCallbacksMixin
from .utils import ensure_iterable

if TYPE_CHECKING:
    from .events import Event
    from .state import State


class TransitionList(AddCallbacksMixin):
    """A list-like container of :ref:`transitions` with callback functions."""

    def __init__(self, transitions: "Iterable[Transition] | None" = None):
        """
        Args:
            transitions: An iterable of `Transition` objects.
                Defaults to `None`.

        """
        self.transitions: List[Transition] = list(transitions) if transitions else []

    def __repr__(self):
        """Return a string representation of the :ref:`TransitionList`."""
        return f"{type(self).__name__}({self.transitions!r})"

    def __or__(self, other: "TransitionList | Iterable"):
        """Return a new :ref:`TransitionList` that combines the transitions of this
        :ref:`TransitionList` with another :ref:`TransitionList` or iterable.

        Args:
            other: Another :ref:`TransitionList` or iterable of :ref:`Transition` objects.

        Returns:
            TransitionList: A new :ref:`TransitionList` object that combines the
                transitions of this :ref:`TransitionList` with `other`.

        """
        return TransitionList(self.transitions).add_transitions(other)

    def _on_event_defined(self, event: str, states: List["State"]):
        self.add_event(event)

        for transition in self.transitions:
            transition.source._on_event_defined(event=event, transition=transition, states=states)

    def add_transitions(self, transition: "Transition | TransitionList | Iterable"):
        """Adds one or more transitions to the :ref:`TransitionList` instance.

        Args:
            transition: A sequence of transitions or a :ref:`TransitionList` instance.

        Returns:
            The updated :ref:`TransitionList` instance.
        """
        if isinstance(transition, TransitionList):
            transition = transition.transitions
        transitions = ensure_iterable(transition)

        for transition in transitions:
            assert isinstance(transition, Transition)  # makes mypy happy
            self.transitions.append(transition)

        return self

    def __getitem__(self, index: int) -> "Transition":
        """Returns the :ref:`transition` at the specified ``index``.

        Args:
            index: The index of the transition.

        Returns:
            The :ref:`transition` at the specified index.
        """
        return self.transitions[index]

    def __len__(self):
        """Returns the number of transitions in the :ref:`TransitionList` instance.

        Returns:
            The number of transitions.
        """
        return len(self.transitions)

    def _add_callback(self, callback, grouper: CallbackGroup, is_event=False, **kwargs):
        for transition in self.transitions:
            list_obj = transition._specs.grouper(grouper)
            list_obj._add_unbounded_callback(
                callback,
                is_event=is_event,
                transitions=self,
                **kwargs,
            )
        return callback

    def add_event(self, event: str):
        """
        Adds an event to all transitions in the :ref:`TransitionList` instance.

        Args:
            event: The name of the event to be added.
        """
        for transition in self.transitions:
            transition.add_event(event)

    @property
    def unique_events(self) -> List["Event"]:
        """
        Returns a list of unique event names across all transitions in the :ref:`TransitionList`
        instance.

        Returns:
            A list of unique event names.
        """
        tmp_ordered_unique_events_as_keys_on_dict = {}
        for transition in self.transitions:
            for event in transition.events:
                tmp_ordered_unique_events_as_keys_on_dict[event] = True

        return list(tmp_ordered_unique_events_as_keys_on_dict.keys())



from celery import Task
from loguru import logger

from genie_flow.session_lock import SessionLockManager


class ProgressLoggingTask(Task):
    """
    This is a celery task that logs the progress of a session.
    It also updates the progress of the session in the session lock manager.
    """
    session_lock_manager: SessionLockManager

    def on_success(self, retval, task_id, args, kwargs):
        session_id: str = args[-3]
        invocation_id: str = args[-1]

        logger.info(
            "Just finished task {task_id} successfully, "
            "for session {session_id} invocation {invocation_id}",
            task_id=task_id,
            session_id=session_id,
            invocation_id=invocation_id,
        )
        logger.debug(
            "Task {task_id} has return value: {retval}",
            task_id=task_id,
            retval=retval,
        )
        self.session_lock_manager.progress_update_done(session_id, invocation_id)

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        session_id: str = args[-3]
        invocation_id: str = args[-1]

        logger.error(
            "Task {task_id} for session {session_id} invocation {invocation_id} "
            "failed with {exc}",
            task_id=task_id,
            session_id=session_id,
            invocation_id=invocation_id,
            exc=exc,
        )
        self.session_lock_manager.progress_update_done(session_id, invocation_id)



from genie_flow_invoker import GenieInvoker


class VerbatimInvoker(GenieInvoker):

    @classmethod
    def from_config(cls, config: dict):
        return cls()

    def invoke(self, content: str) -> str:
        """
        Invokes the verbatim invoker that just copies the content as a result.

        :param content: Any text content
        :returns: the `str` version of the content
        """
        return str(content)



from typing import Type, Optional, Literal

import redis_lock
from loguru import logger
from redis import Redis

from genie_flow.genie import GenieModel
from genie_flow.model.persistence import Persistence, PersistenceLevel
from genie_flow.model.secondary_store import SecondaryStore
from genie_flow.model.versioned import VersionedModel
from genie_flow.model.user import User
from genie_flow.mongo import retrieve_model
from genie_flow.utils import get_class_from_fully_qualified_name, get_fully_qualified_name_from_class


StoreType = Literal["object", "secondary", "lock", "progress"]


class SessionLockManager:

    class ModelContextManager:

        def __init__(
                self,
                lock_manager: "SessionLockManager",
                session_id: str,
                model_class: Type[GenieModel]
        ):
            """
            A Context Manager for obtaining a locked `GenieModel`. This context manager obtains
            a lock for a particular session, then retrieves the GenieModel from store. Upon the
            exit of the context, that GenieModel is stored back into store and the lock is released.
            """
            self.lock_manager = lock_manager
            self.model_class = model_class
            self.session_id = session_id
            self.lock = lock_manager._create_lock_for_session(session_id)
            self.model: Optional[GenieModel] = None

        def __enter__(self) -> GenieModel:
            self.lock.acquire()
            self.model = self.lock_manager._retrieve_model(self.session_id, self.model_class)
            return self.model

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.lock_manager._store_model(self.model)
            self.lock.release()

    def __init__(
        self,
        redis_object_store: Redis,
        redis_lock_store: Redis,
        redis_progress_store: Redis,
        object_expiration_seconds: int,
        lock_expiration_seconds: int,
        progress_expiration_seconds: int,
        compression: bool,
        application_prefix: str,
    ):
        """
        The `SessionLockManager` manages the session lock as well as the retrieval and persisting
        of model objects. When changes are (expected to be) made to the model of a particular
        session, this manager deals with locking multithreaded access to it when it retrieves
        from Store, and before it gets written back to it.
        :param redis_object_store: The Redis object store
        :param redis_lock_store: The Redis lock store
        :param redis_progress_store: The Redis progress store
        :param object_expiration_seconds: The expiration time for objects in seconds
        :param lock_expiration_seconds: The expiration time of the lock in seconds
        :param progress_expiration_seconds: The expiration time of the progress object in seconds
        :param compression: Whether or not to compress the model when persisting
        :param application_prefix: The application prefix used to create the key for an object
        """
        self.redis_object_store = redis_object_store
        self.redis_lock_store = redis_lock_store
        self.redis_progress_store = redis_progress_store
        self.object_expiration_seconds = object_expiration_seconds
        self.lock_expiration_seconds = lock_expiration_seconds
        self.progress_expiration_seconds = progress_expiration_seconds
        self.compression = compression
        self.application_prefix = application_prefix
        self.update_set_key="update:session"
       
    def _create_lock_for_session(self, session_id: str) -> redis_lock.Lock:
        """
        Retrieve the lock for the object for the given `session_id`. This ensures that only
        one process will have access to the model and potentially make changes to it.
        This lock can function as a context manager. See the documentation of `redis_lock.Lock`
        :param session_id: The session id that the object in question belongs to
        """
        lock = redis_lock.Lock(
            self.redis_lock_store,
            name=session_id,
            expire=self.lock_expiration_seconds,
            auto_renewal=True,
        )
        return lock

    def _create_key(
            self,
            store: StoreType,
            model: GenieModel | type[GenieModel] | None,
            *args: str,
    ) -> str:
        key = f"{self.application_prefix}:{store}"
        if model is None:
            model_indicator = ""
        elif isinstance(model, GenieModel):
            model_indicator = model.__class__.__name__
        else:
            model_indicator = model.__name__
        return key + f":{model_indicator}:" + ":".join([arg for arg in args if arg is not None])

    def _retrieve_secondary_storage(
            self,
            session_id: str,
            model_cls: Type[GenieModel],
    ) -> SecondaryStore:
        """
        Retrieve the secondary storage for the given session id and model class.
        Not protected by a lock, and the user should ensure that no other process is accessing
        the model's secondary storage at the same time.
        :param session_id: the session id for which to retrieve the secondary storage
        :param model_cls: the model class for which to retrieve the secondary storage
        :return: a newly instantiated SecondaryStore object for the given session id and model class
        """
        secondary_key = self._create_key("secondary", model_cls, session_id)
        serialized_values = self.redis_object_store.hgetall(secondary_key)
        return SecondaryStore.from_serialized(serialized_values)

    def _retrieve_model(self, session_id: str, model_class: Type[GenieModel]) -> GenieModel:
        """
        Retrieve the GenieModel for the object for the given `session_id`. This retrieval is
        not protected by a lock, and the user should ensure that no other process is accessing
        the model at the same time.
        :param session_id: the session id that the object in question belongs to
        :param model_class: the GenieModel class to retrieve
        :return: a retrieved GenieModel object for the given `session_id`
        """
        model_key = self._create_key("object", model_class, session_id)
        payload = self.redis_object_store.get(model_key)
        if payload is None:
            logger.error("No model with id {session_id} found in object store, trying mongodb", session_id=session_id)
            try:
                mongo_data = retrieve_model(session_id)
                payload = mongo_data['model']
            except:
                raise KeyError(f"No model with id {session_id}")

        model = model_class.deserialize(payload)
        model.secondary_storage = self._retrieve_secondary_storage(session_id, model_class)
        return model

    def get_model(self, session_id: str, model_class: str | Type[GenieModel]) -> GenieModel:
        """
        Retrieve the GenieModel for the object for the given `session_id`. This retrieval is
        done inside a locked context, so no writing of the GenieModel can happen when this
        retrieval is done.

        :param session_id: The session id that the object in question belongs to
        :param model_class: The GenieModel class to retrieve or the fqn of the class
        :return: The GenieModel object for the given `session_id`
        """
        if isinstance(model_class, str):
            model_class = get_class_from_fully_qualified_name(model_class)
        with self._create_lock_for_session(session_id):
            return self._retrieve_model(session_id, model_class)

    def _store_secondary_storage(self, model: GenieModel):
        """
        Store the secondary storage values from the given Genie Model.
        Will only persist properties that have not yet been stored before.

        :param model: The Genie Model containing the secondary store to persist
        """
        secondary_key = self._create_key("secondary", model, model.session_id)

        if model.secondary_storage.has_unpersisted_values:
            unpersisted_serialized = model.secondary_storage.unpersisted_serialized(
                self.compression,
            )
            logger.debug(
                "Writing unpersisted field(s) [{field_list}] to secondary storage "
                "for session {session_id}",
                field_list=", ".join(unpersisted_serialized.keys()),
                session_id=model.session_id,
            )
            self.redis_object_store.hset(secondary_key, mapping=unpersisted_serialized)
            model.secondary_storage.mark_persisted(unpersisted_serialized.keys())

        if model.secondary_storage.has_deleted_values:
            deleted_fields = model.secondary_storage.deleted_keys
            logger.debug(
                "Removing deleted field(s) [{fields}] from secondary storage "
                "for session {session_id}",
                fields=", ".join(deleted_fields),
                session_id=model.session_id,
            )
            self.redis_object_store.hdel(secondary_key, *deleted_fields)

    def _store_model(self, model: GenieModel):
        """
        Underlying logic of writing a Genie Model to the object store.
        No locking happens in this method, so user is responsible for
        making sure no parallel reading or writing is done.

        :param model: the GenieModel to store
        """
        model_key = self._create_key("object", model, model.session_id)
        logger.debug(
            "Storing model for session {session_id} in object store",
            session_id=model.session_id,
        )

        self._store_secondary_storage(model)

        self.redis_object_store.set(
            model_key,
            model.serialize(self.compression, exclude={"secondary_storage"}),
            ex=self.object_expiration_seconds,
        )
        model_fqn = get_fully_qualified_name_from_class(model)
        if "persistence" not in  model.secondary_storage or \
            model.secondary_storage["persistence"].level == PersistenceLevel.LONG_TERM_PERSISTENCE:
            self.redis_object_store.sadd(
                self.update_set_key,
                f"{model_fqn}:{model.session_id}"
            )
            logger.debug(f"{self.redis_object_store.smembers(self.update_set_key)}")

    def store_model(self, model: GenieModel):
        """
        Store a model into the object store. Also stores the secondary storage of the model.
        Storing happens inside a locked context, so no reading or writing of the GenieModel
        can happen when this method is executing.

        :param model: the object to store
        """
        with self._create_lock_for_session(model.session_id):
            self._store_model(model)

    def get_locked_model(
            self,
            session_id: str,
            model_class: str | Type[GenieModel],
    ) -> ModelContextManager:
        if isinstance(model_class, str):
            model_class = get_class_from_fully_qualified_name(model_class)
        return SessionLockManager.ModelContextManager(self, session_id, model_class)

    @staticmethod
    def _create_field_key(field: str, invocation_id: str) -> str:
        return f"{invocation_id}:{field}"

    def _create_existing_progress_key(
            self,
            action: str,
            session_id: str,
            invocation_id: str
    ) -> str:
        progress_key = self._create_key("progress", None, session_id)
        if not self.redis_progress_store.exists(progress_key):
            logger.error(
                "Action {action} but no progress record for session {session_id}",
                action=action,
                session_id=session_id,
            )
            raise KeyError("No progress record for session")
        if not self.redis_progress_store.hexists(
            progress_key,
            self._create_field_key("todo", invocation_id),
        ):
            logger.error(
                "Action {action} but no progress record for session {session_id} "
                "and invocation {invocation_id}",
                action=action,
                session_id=session_id,
                invocation_id=invocation_id,
            )
            raise KeyError("No progress record for session and invocation")

        logger.info(
            "Starting {action} for session {session_id} and invocation {invocation_id}",
            action=action,
            session_id=session_id,
            invocation_id=invocation_id,
        )
        return progress_key

    def progress_start(
            self,
            session_id: str,
            invocation_id: str,
            nr_tasks_todo: int,
    ):
        progress_key = self._create_key(
            "progress",
            None,
            session_id,
        )
        if (
            self.redis_progress_store.exists(progress_key) > 0
            and self.redis_progress_store.hexists(
                progress_key,
                self._create_field_key("todo", invocation_id),
            )
        ):
            logger.error(
                "Progress record for session {session_id} and invocation {invocation_id} already exists",
                session_id=session_id,
                invocation_id=invocation_id,
            )
            raise ValueError("Progress record already exists for session and invocation")

        logger.info(
            "Starting progress record for session {session_id} and invocation {invocation_id}, "
            "with {nr_todo} tasks",
            session_id=session_id,
            invocation_id=invocation_id,
            nr_todo=nr_tasks_todo,
        )
        self.redis_progress_store.hset(
            progress_key,
            mapping={
                self._create_field_key("todo", invocation_id): nr_tasks_todo,
                self._create_field_key("done", invocation_id): 0,
                self._create_field_key("tombstone", invocation_id): "f",
            },
        )

    def progress_exists(self, session_id: str, invocation_id: Optional[str] = None) -> bool:
        progress_key = self._create_key(
            "progress",
            None,
            session_id,
        )
        if invocation_id is None:
            return self.redis_progress_store.exists(progress_key) == 1

        return self.redis_progress_store.hexists(
            progress_key,
            self._create_field_key("todo", invocation_id),
        )

    def progress_update_todo(
            self,
            session_id: str,
            invocation_id: str,
            nr_increase: int,
    ) -> int:
        progress_key = self._create_existing_progress_key(
            "Update To Do Count",
            session_id,
            invocation_id,
        )
        new_todo = self.redis_progress_store.hincrby(
            progress_key,
            self._create_field_key("todo", invocation_id),
            nr_increase,
        )
        logger.debug(
            "New: {new_todo} tasks to do for session {session_id}, invocation {invocation_id}",
            new_todo=new_todo,
            session_id=session_id,
            invocation_id=invocation_id,
        )
        return new_todo

    def progress_update_done(
            self,
            session_id: str,
            invocation_id: str,
            nr_done: int = 1,
    ) -> int:
        progress_key = self._create_existing_progress_key(
            "Update Done Count",
            session_id,
            invocation_id,
        )

        new_done = self.redis_progress_store.hincrby(
            progress_key,
            self._create_field_key("done", invocation_id),
            nr_done,
        )
        logger.debug(
            "New: {new_done} tasks done for session {session_id}, invocation {invocation_id}",
            new_done=new_done,
            session_id=session_id,
            invocation_id=invocation_id,
        )
        tombstone, todo_str = self.redis_progress_store.hmget(
            progress_key,
            [
                self._create_field_key("tombstone", invocation_id),
                self._create_field_key("todo", invocation_id),
            ],
        )
        todo = int(todo_str)
        
        if tombstone == b"t":
            if todo > new_done:
                logger.warning(
                    "Got an update for session {session_id} and invocation {invocation_id} "
                    "with a tombstoned progress record, with tasks to do {todo} > {done} done",
                    session_id=session_id,
                    invocation_id=invocation_id,
                    todo=todo,
                    done=new_done,
                )
            logger.info(
                "Removing progress record for session {session_id} and invocation {invocation_id}",
                session_id=session_id,
                invocation_id=invocation_id,
            )
            self.redis_progress_store.hdel(
                progress_key,
                self._create_field_key("todo", invocation_id),
                self._create_field_key("done", invocation_id),
                self._create_field_key("tombstone", invocation_id),
            )
        elif new_done >= todo:
            logger.warning(
                "Progress record for session {session_id} and invocation {invocation_id} "
                "indicates done {done} tasks >= {todo} tasks to do",
                session_id=session_id,
                invocation_id=invocation_id,
                done=new_done,
                todo=todo,
            )
        return todo - new_done

    def progress_tombstone(self, session_id: str, invocation_id: str):
        progress_key = self._create_existing_progress_key(
            "Tombstone Progress Record",
            session_id,
            invocation_id,
        )
        self.redis_progress_store.hset(
            progress_key,
            self._create_field_key("tombstone", invocation_id),
            "t"
        )

    def progress_status(self, session_id: str) -> tuple[int, int]:
        progress_key = self._create_key("progress", None, session_id)
        if not self.redis_progress_store.exists(progress_key):
            return 0, 0

        invocations_to_ignore = set()
        field_values = self.redis_progress_store.hgetall(progress_key)
        for field_name, value in field_values.items():
            if field_name.endswith(b"tombstone") and value == b"t":
                invocations_to_ignore.add(field_name.split(b":")[0])

        todo, done = 0, 0
        for field_name, value in field_values.items():
            invocation_id, field = field_name.split(b":")
            if invocation_id in invocations_to_ignore:
                continue
            if field == b"todo":
                todo += int(value)
            elif field == b"done":
                done += int(value)

        logger.debug(
            "Found there are {todo} - {done} = {nr_left} tasks left to do for "
            "session {session_id}",
            todo=todo,
            done=done,
            nr_left=todo - done,
            session_id=session_id,
        )
        return todo, done



from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from weakref import ref

from .callbacks import CallbackGroup
from .callbacks import CallbackPriority
from .callbacks import CallbackSpecList
from .exceptions import StateMachineError
from .i18n import _
from .transition import Transition
from .transition_list import TransitionList

if TYPE_CHECKING:
    from .statemachine import StateMachine


class _TransitionBuilder:
    def __init__(self, state: "State"):
        self._state = state

    def itself(self, **kwargs):
        return self.__call__(self._state, **kwargs)

    def __call__(self, *states: "State", **kwargs):
        raise NotImplementedError


class _ToState(_TransitionBuilder):
    def __call__(self, *states: "State", **kwargs):
        transitions = TransitionList(Transition(self._state, state, **kwargs) for state in states)
        self._state.transitions.add_transitions(transitions)
        return transitions


class _FromState(_TransitionBuilder):
    def any(self, **kwargs):
        """Create transitions from all non-final states (reversed)."""
        return self.__call__(AnyState(), **kwargs)

    def __call__(self, *states: "State", **kwargs):
        transitions = TransitionList()
        for origin in states:
            transition = Transition(origin, self._state, **kwargs)
            origin.transitions.add_transitions(transition)
            transitions.add_transitions(transition)
        return transitions


class State:
    """
    A State in a :ref:`StateMachine` describes a particular behavior of the machine.
    When we say that a machine is “in” a state, it means that the machine behaves
    in the way that state describes.

    Args:
        name: A human-readable representation of the state. Default is derived
            from the name of the variable assigned to the state machine class.
            The name is derived from the id using this logic::

                name = id.replace("_", " ").capitalize()

        value: A specific value to the storage and retrieval of states.
            If specified, you can use It to map a more friendly representation to a low-level
            value.
        initial: Set ``True`` if the ``State`` is the initial one. There must be one and only
            one initial state in a statemachine. Defaults to ``False``.
        final: Set ``True`` if represents a final state. A machine can have
            optionally many final states. Final states have no :ref:`transition` starting from It.
            Defaults to ``False``.
        enter: One or more callbacks assigned to be executed when the state is entered.
            See :ref:`actions`.
        exit: One or more callbacks assigned to be executed when the state is exited.
            See :ref:`actions`.

    State is a core component on how this library implements an expressive API to declare
    StateMachines.

    >>> from statemachine import State

    Given a few states...

    >>> draft = State("Draft", initial=True)

    >>> producing = State("Producing")

    >>> closed = State('Closed', final=True)

    Transitions are declared using the :func:`State.to` or :func:`State.from_` (reversed) methods.

    >>> draft.to(producing)
    TransitionList([Transition(State('Draft', ...

    The result is a :ref:`TransitionList`.
    Don't worry about this internal class.
    But the good thing is that it implements the ``OR`` operator to combine transitions,
    so you can use the ``|`` syntax to compound a list of transitions and assign
    to the same event.

    You can declare all transitions for a state in one single line ...

    >>> transitions = draft.to(draft) | producing.to(closed)

    ... and you can append additional transitions for a state to previous definitions.

    >>> transitions |= closed.to(draft)

    >>> [(t.source.name, t.target.name) for t in transitions]
    [('Draft', 'Draft'), ('Producing', 'Closed'), ('Closed', 'Draft')]

    There are handy shortcuts that you can use to express this same set of transitions.

    The first one, ``draft.to(draft)``, is also called a :ref:`self-transition`, and can be
    expressed using an alternative syntax:

    >>> draft.to.itself()
    TransitionList([Transition(State('Draft', ...

    You can even pass a list of target states to declare at once all transitions starting
    from the same state.

    >>> transitions = draft.to(draft, producing, closed)

    >>> [(t.source.name, t.target.name) for t in transitions]
    [('Draft', 'Draft'), ('Draft', 'Producing'), ('Draft', 'Closed')]

    Sometimes it's easier to use the :func:`State.from_` method:

    >>> transitions = closed.from_(draft, producing, closed)

    >>> [(t.source.name, t.target.name) for t in transitions]
    [('Draft', 'Closed'), ('Producing', 'Closed'), ('Closed', 'Closed')]

    """

    def __init__(
        self,
        name: str = "",
        value: Any = None,
        initial: bool = False,
        final: bool = False,
        enter: Any = None,
        exit: Any = None,
    ):
        self.name = name
        self.value = value
        self._initial = initial
        self._final = final
        self._id: str = ""
        self.transitions = TransitionList()
        self._specs = CallbackSpecList()
        self.enter = self._specs.grouper(CallbackGroup.ENTER).add(
            enter, priority=CallbackPriority.INLINE
        )
        self.exit = self._specs.grouper(CallbackGroup.EXIT).add(
            exit, priority=CallbackPriority.INLINE
        )

    def __eq__(self, other):
        return isinstance(other, State) and self.name == other.name and self.id == other.id

    def __hash__(self):
        return hash(repr(self))

    def _setup(self):
        self.enter.add("on_enter_state", priority=CallbackPriority.GENERIC, is_convention=True)
        self.enter.add(f"on_enter_{self.id}", priority=CallbackPriority.NAMING, is_convention=True)
        self.exit.add("on_exit_state", priority=CallbackPriority.GENERIC, is_convention=True)
        self.exit.add(f"on_exit_{self.id}", priority=CallbackPriority.NAMING, is_convention=True)

    def _on_event_defined(self, event: str, transition: Transition, states: List["State"]):
        """Called by statemachine factory when an event is defined having a transition
        starting from this state.
        """
        pass

    def __repr__(self):
        return (
            f"{type(self).__name__}({self.name!r}, id={self.id!r}, value={self.value!r}, "
            f"initial={self.initial!r}, final={self.final!r})"
        )

    def __str__(self):
        return self.name

    def __get__(self, machine, owner):
        if machine is None:
            return self
        return self.for_instance(machine=machine, cache=machine._states_for_instance)

    def __set__(self, instance, value):
        raise StateMachineError(
            _("State overriding is not allowed. Trying to add '{}' to {}").format(value, self.id)
        )

    def for_instance(self, machine: "StateMachine", cache: Dict["State", "State"]) -> "State":
        if self not in cache:
            cache[self] = InstanceState(self, machine)

        return cache[self]

    @property
    def id(self) -> str:
        return self._id

    def _set_id(self, id: str):
        self._id = id
        if self.value is None:
            self.value = id
        if not self.name:
            self.name = self._id.replace("_", " ").capitalize()

    @property
    def to(self) -> _ToState:
        """Create transitions to the given target states."""
        return _ToState(self)

    @property
    def from_(self) -> _FromState:
        """Create transitions from the given target states (reversed)."""
        return _FromState(self)

    @property
    def initial(self):
        return self._initial

    @property
    def final(self):
        return self._final


class InstanceState(State):
    """ """

    def __init__(
        self,
        state: State,
        machine: "StateMachine",
    ):
        self._state = ref(state)
        self._machine = ref(machine)

    @property
    def name(self):
        return self._state().name

    @property
    def value(self):
        return self._state().value

    @property
    def transitions(self):
        return self._state().transitions

    @property
    def enter(self):
        return self._state().enter

    @property
    def exit(self):
        return self._state().exit

    def __eq__(self, other):
        return self._state() == other

    def __hash__(self):
        return hash(repr(self._state()))

    def __repr__(self):
        return repr(self._state())

    @property
    def initial(self):
        return self._state()._initial

    @property
    def final(self):
        return self._state()._final

    @property
    def id(self) -> str:
        return (self._state() or self)._id

    @property
    def is_active(self):
        return self._machine().current_state == self


class AnyState(State):
    """A special state that works as a "ANY" placeholder.

    It is used as the "From" state of a transtion,
    until the state machine class is evaluated.
    """

    def _on_event_defined(self, event: str, transition: Transition, states: List[State]):
        for state in states:
            if state.final:
                continue
            new_transition = transition._copy_with_args(source=state, event=event)

            state.transitions.add_transitions(new_transition)



#!/usr/bin/env python
#
# Copyright (c) 2011, Andres Moreira <andres@andresmoreira.com>
#               2011, Felipe Cruz <felipecruz@loogica.net>
#               2012, JT Olds <jt@spacemonkey.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the authors nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANDRES MOREIRA BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""python-snappy

Python library for the snappy compression library from Google.
Expected usage like:

    import snappy

    compressed = snappy.compress("some data")
    assert "some data" == snappy.uncompress(compressed)

"""
from __future__ import absolute_import

import struct

import cramjam

_CHUNK_MAX = 65536
_STREAM_TO_STREAM_BLOCK_SIZE = _CHUNK_MAX
_STREAM_IDENTIFIER = b"sNaPpY"
_IDENTIFIER_CHUNK = 0xff
_STREAM_HEADER_BLOCK = b"\xff\x06\x00\x00sNaPpY"

_compress = cramjam.snappy.compress_raw
_uncompress = cramjam.snappy.decompress_raw


class UncompressError(Exception):
    pass


def isValidCompressed(data):
    if isinstance(data, str):
        data = data.encode('utf-8')

    ok = True
    try:
        decompress(data)
    except UncompressError as err:
        ok = False
    return ok


def compress(data, encoding='utf-8'):
    if isinstance(data, str):
        data = data.encode(encoding)

    return bytes(_compress(data))

def uncompress(data, decoding=None):
    if isinstance(data, str):
        raise UncompressError("It's only possible to uncompress bytes")
    try:
        out = bytes(_uncompress(data))
    except cramjam.DecompressionError as err:
        raise UncompressError from err
    if decoding:
        return out.decode(decoding)
    return out


decompress = uncompress

class StreamCompressor():

    """This class implements the compressor-side of the proposed Snappy framing
    format, found at

        http://code.google.com/p/snappy/source/browse/trunk/framing_format.txt
            ?spec=svn68&r=71

    This class matches the interface found for the zlib module's compression
    objects (see zlib.compressobj), but also provides some additions, such as
    the snappy framing format's ability to intersperse uncompressed data.

    Keep in mind that this compressor object does no buffering for you to
    appropriately size chunks. Every call to StreamCompressor.compress results
    in a unique call to the underlying snappy compression method.
    """

    def __init__(self):
        self.c = cramjam.snappy.Compressor()

    def add_chunk(self, data: bytes, compress=None):
        """Add a chunk, returning a string that is framed and compressed. 
        
        Outputs a single snappy chunk; if it is the very start of the stream,
        will also contain the stream header chunk.
        """
        self.c.compress(data)
        return self.flush()

    compress = add_chunk

    def flush(self):
        return bytes(self.c.flush())

    def copy(self):
        """This method exists for compatibility with the zlib compressobj.
        """
        return self


class StreamDecompressor():

    """This class implements the decompressor-side of the proposed Snappy
    framing format, found at

        http://code.google.com/p/snappy/source/browse/trunk/framing_format.txt
            ?spec=svn68&r=71

    This class matches a subset of the interface found for the zlib module's
    decompression objects (see zlib.decompressobj). Specifically, it currently
    implements the decompress method without the max_length option, the flush
    method without the length option, and the copy method.
    """
    def __init__(self):
        self.c = cramjam.snappy.Decompressor()
        self.remains = None
    
    @staticmethod
    def check_format(fin):
        """Does this stream start with a stream header block?

        True indicates that the stream can likely be decoded using this class.
        """
        try:
            return fin.read(len(_STREAM_HEADER_BLOCK)) == _STREAM_HEADER_BLOCK
        except:
            return False

    def decompress(self, data: bytes):
        """Decompress 'data', returning a string containing the uncompressed
        data corresponding to at least part of the data in string. This data
        should be concatenated to the output produced by any preceding calls to
        the decompress() method. Some of the input data may be preserved in
        internal buffers for later processing.
        """
        if self.remains:
            data = self.remains + data
            self.remains = None
        if not data.startswith(_STREAM_HEADER_BLOCK):
            data = _STREAM_HEADER_BLOCK + data
        ldata = len(data)
        bsize = len(_STREAM_HEADER_BLOCK)
        if bsize + 4 > ldata:
            # not even enough for one block
            self.remains = data
            return b""
        while True:
            this_size = int.from_bytes(data[bsize + 1: bsize + 4], "little") + 4
            if bsize == ldata:
                # ended on a block boundary
                break
            if this_size + bsize > ldata:
                # last block incomplete
                self.remains = data[bsize:]
                data = data[:bsize]
                break
            bsize += this_size
        self.c.decompress(data)
        return self.flush()

    def flush(self):
        return bytes(self.c.flush())

    def copy(self):
        return self


class HadoopStreamCompressor():
    def add_chunk(self, data: bytes, compress=None):
        """Add a chunk, returning a string that is framed and compressed. 
        
        Outputs a single snappy chunk; if it is the very start of the stream,
        will also contain the stream header chunk.
        """
        cdata = _compress(data)
        return b"".join((len(data).to_bytes(4, "big"), len(cdata).to_bytes(4, "big"), cdata))

    compress = add_chunk

    def flush(self):
        # never maintains a buffer
        return b""

    def copy(self):
        """This method exists for compatibility with the zlib compressobj.
        """
        return self


class HadoopStreamDecompressor():
    def __init__(self):
        self.remains = b""
    
    @staticmethod
    def check_format(fin):
        """Does this look like a hadoop snappy stream?
        """
        try:
            from snappy.snappy_formats import check_unframed_format
            size = fin.seek(0, 2)
            fin.seek(0)
            assert size >= 8

            chunk_length = int.from_bytes(fin.read(4), "big")
            assert chunk_length < size
            fin.read(4)
            return check_unframed_format(fin)
        except:
            return False

    def decompress(self, data: bytes):
        """Decompress 'data', returning a string containing the uncompressed
        data corresponding to at least part of the data in string. This data
        should be concatenated to the output produced by any preceding calls to
        the decompress() method. Some of the input data may be preserved in
        internal buffers for later processing.
        """
        if self.remains:
            data = self.remains + data
            self.remains = None
        if len(data) < 8:
            self.remains = data
            return b""
        out = []
        while True:
            chunk_length = int.from_bytes(data[4:8], "big")
            if len(data) < 8 + chunk_length:
                self.remains = data
                break
            out.append(_uncompress(data[8:8 + chunk_length]))
            data = data[8 + chunk_length:]
        return b"".join(out)

    def flush(self):
        return b""

    def copy(self):
        return self
        


def stream_compress(src,
                    dst,
                    blocksize=_STREAM_TO_STREAM_BLOCK_SIZE,
                    compressor_cls=StreamCompressor):
    """Takes an incoming file-like object and an outgoing file-like object,
    reads data from src, compresses it, and writes it to dst. 'src' should
    support the read method, and 'dst' should support the write method.

    The default blocksize is good for almost every scenario.
    """
    compressor = compressor_cls()
    while True:
        buf = src.read(blocksize)
        if not buf: break
        buf = compressor.add_chunk(buf)
        if buf: dst.write(buf)


def stream_decompress(src,
                      dst,
                      blocksize=_STREAM_TO_STREAM_BLOCK_SIZE,
                      decompressor_cls=StreamDecompressor,
                      start_chunk=None):
    """Takes an incoming file-like object and an outgoing file-like object,
    reads data from src, decompresses it, and writes it to dst. 'src' should
    support the read method, and 'dst' should support the write method.

    The default blocksize is good for almost every scenario.
    :param decompressor_cls: class that implements `decompress` method like
        StreamDecompressor in the module
    :param start_chunk: start block of data that have already been read from
        the input stream (to detect the format, for example)
    """
    decompressor = decompressor_cls()
    while True:
        if start_chunk:
            buf = start_chunk
            start_chunk = None
        else:
            buf = src.read(blocksize)
            if not buf: break
        buf = decompressor.decompress(buf)
        if buf: dst.write(buf)
    decompressor.flush()  # makes sure the stream ended well


def hadoop_stream_decompress(
    src,
    dst,
    blocksize=_STREAM_TO_STREAM_BLOCK_SIZE,
):
    c = HadoopStreamDecompressor()
    while True:
        data = src.read(blocksize)
        if not data:
            break
        buf = c.decompress(data)
        if buf:
            dst.write(buf)
    dst.flush()


def hadoop_stream_compress(
    src,
    dst,
    blocksize=_STREAM_TO_STREAM_BLOCK_SIZE,
):
    c = HadoopStreamCompressor()
    while True:
        data = src.read(blocksize)
        if not data:
            break
        buf = c.compress(data)
        if buf:
            dst.write(buf)
    dst.flush()


def raw_stream_decompress(src, dst):
    data = src.read()
    dst.write(decompress(data))


def raw_stream_compress(src, dst):
    data = src.read()
    dst.write(compress(data))



from abc import ABC
from typing import Optional

from pydantic import BaseModel, Field

from genie_flow.genie import GenieModel
from genie_flow.model.user import User
from genie_flow.model.versioned import VersionedModel


class GenieMessage(BaseModel, ABC):
    """
    The base class for any message that is exchanged through the endpoints.
    """

    session_id: str = Field(
        description="the Session ID associated with the interface message"
    )


class GenieMessageWithActions(GenieMessage, ABC):
    """
    The `GenieMessage` that also carries a list of events that can be sent, given the current
    state of the model instance.
    """

    next_actions: list[str] = Field(
        default_factory=list,
        description="A list of actions that the user can send to evolve to the next state",
    )


class AIStatusResponse(GenieMessageWithActions):
    """
    The `GenieMessage` that represents the status of a background task that may run for the
    model object that belongs to a session.
    """

    ready: bool = Field(description="indicated if the response is ready")


class AIProgressResponse(BaseModel):
    total_number_of_subtasks: int = Field()
    number_of_subtasks_executed: int = Field()

    @property
    def percent_completed(self) -> float:
        if self.number_of_subtasks_executed == 0:
            return 100.0

        return 100.0 * float(self.number_of_subtasks_executed) / float(self.total_number_of_subtasks)


class AIResponse(GenieMessageWithActions):
    """
    The response that results from sending an event and the model instance's state machine
    to transition into the next state.
    """

    error: Optional[str] = Field(
        default=None,
        description="A potential error message",
    )
    response: Optional[str] = Field(
        default=None,
        description="The text response from the AI service",
    )
    progress: Optional[AIProgressResponse] = Field(
        default=None,
        description="The progress of any background process execution",
    )


class EventInput(GenieMessage):
    """
    An event with accompanying information that is sent to progress the state machine that
    belongs to a session.
    """

    event: str = Field(description="The name of the event that is triggered")
    event_input: str = Field(description="The string input that belongs to this event")


class SessionStartRequest(BaseModel):
    event: str = Field(description="The name of the event that is triggered")
    event_input: str = Field(description="The string input that belongs to this event")
    user_info: Optional[User] = Field(
        default=None,
        description="The user information that is associated with the session",
    )



from .event import Event
from .state import State
from .statemachine import StateMachine

__author__ = """Fernando Macedo"""
__email__ = "fgmacedo@gmail.com"
__version__ = "2.5.0"

__all__ = ["StateMachine", "State", "Event"]



from enum import Enum
from typing import Dict  # deprecated since 3.9: https://peps.python.org/pep-0585/
from typing import Type

from .state import State
from .utils import ensure_iterable

EnumType = Type[Enum]


class States:
    """
    A class representing a collection of :ref:`State` objects.

    Helps creating :ref:`StateMachine`'s :ref:`state` definitions from other
    sources, like an ``Enum`` class, using :meth:`States.from_enum`.

    >>> states_def = [('open', {'initial': True}), ('closed', {'final': True})]

    >>> from statemachine import StateMachine
    >>> class SM(StateMachine):
    ...
    ...     states = States({
    ...         name: State(**params) for name, params in states_def
    ...     })
    ...
    ...     close = states.open.to(states.closed)

    And states can be used as usual.

    >>> sm = SM()
    >>> sm.send("close")
    >>> sm.current_state.id
    'closed'

    """

    def __init__(self, states: "Dict[str, State] | None" = None) -> None:
        """
        Initializes a new instance of the States class.

        Args:
            states: A dictionary mapping keys as ``State.id`` and values :ref:`state` instances.

        Returns:
            None.
        """
        self._states: Dict[str, State] = states if states is not None else {}

    def __repr__(self):
        return f"{list(self)}"

    def __eq__(self, other):
        return list(self) == list(other)

    def __getattr__(self, name: str):
        if name in self._states:
            return self._states[name]
        raise AttributeError(f"{name} not found in {self.__class__.__name__}")

    def __len__(self):
        return len(self._states)

    def __iter__(self):
        return iter(self._states.values())

    def append(self, state):
        self._states[state.id] = state

    def items(self):
        """
        Returns a view object of the states, with pairs of ``(State.id, State)``.

        Args:
            None.

        Returns:
            A view object of the items in the the instance.
        """
        return self._states.items()

    @classmethod
    def from_enum(cls, enum_type: EnumType, initial, final=None, use_enum_instance: bool = False):
        """
        Creates a new instance of the ``States`` class from an enumeration.

        Consider an ``Enum`` type that declares our expected states:

        >>> class Status(Enum):
        ...     pending = 1
        ...     completed = 2

        A :ref:`StateMachine` that uses this enum can be declared as follows:

        >>> from statemachine import StateMachine
        >>> class ApprovalMachine(StateMachine):
        ...
        ...     _ = States.from_enum(Status, initial=Status.pending, final=Status.completed)
        ...
        ...     finish = _.pending.to(_.completed)
        ...
        ...     def on_enter_completed(self):
        ...         print("Completed!")

        .. tip::
            When you assign the result of ``States.from_enum`` to a class-level variable in your
            :ref:`StateMachine`, you're all set. You can use any name for this variable. In this
            example, we used ``_`` to show that the name doesn't matter. The metaclass will inspect
            the variable of type :ref:`States (class)` and automatically assign the inner
            :ref:`State` instances to the state machine.


        Everything else is similar, the ``Enum`` is only used to declare the :ref:`State`
        instances.

        >>> sm = ApprovalMachine()

        >>> sm.pending.is_active
        True

        >>> sm.send("finish")
        Completed!

        >>> sm.completed.is_active
        True

        >>> sm.current_state_value
        2

        If you need to use the enum instance as the state value, you can set the
        ``use_enum_instance=True``:

        >>> states = States.from_enum(Status, initial=Status.pending, use_enum_instance=True)
        >>> states.completed.value
        <Status.completed: 2>

        .. deprecated:: 2.3.3

            On the next major release, ``use_enum_instance=True`` will be the default.

        Args:
            enum_type: An enumeration containing the states of the machine.
            initial: The initial state of the machine.
            final: A set of final states of the machine.
            use_enum_instance: If ``True``, the value of the state will be the enum item instance,
                otherwise the enum item value. Defaults to ``False``.

        Returns:
            A new instance of the :ref:`States (class)`.
        """
        final_set = set(ensure_iterable(final))
        return cls(
            {
                e.name: State(
                    value=(e if use_enum_instance else e.value),
                    initial=e is initial,
                    final=e in final_set,
                )
                for e in enum_type
            }
        )



from enum import Enum

from pydantic import Field

from genie_flow.model.versioned import VersionedModel


class PersistenceLevel(Enum):
    EPHEMERAL = 0
    LONG_TERM_PERSISTENCE = 1


class Persistence(VersionedModel):
    level: PersistenceLevel = Field(
        description="the persistence level of the object"
    )



from __future__ import annotations

import re
from typing import Callable, Iterable, NamedTuple, Sequence

from prompt_toolkit.document import Document
from prompt_toolkit.filters import FilterOrBool, to_filter
from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples

from .base import CompleteEvent, Completer, Completion
from .word_completer import WordCompleter

__all__ = [
    "FuzzyCompleter",
    "FuzzyWordCompleter",
]


class FuzzyCompleter(Completer):
    """
    Fuzzy completion.
    This wraps any other completer and turns it into a fuzzy completer.

    If the list of words is: ["leopard" , "gorilla", "dinosaur", "cat", "bee"]
    Then trying to complete "oar" would yield "leopard" and "dinosaur", but not
    the others, because they match the regular expression 'o.*a.*r'.
    Similar, in another application "djm" could expand to "django_migrations".

    The results are sorted by relevance, which is defined as the start position
    and the length of the match.

    Notice that this is not really a tool to work around spelling mistakes,
    like what would be possible with difflib. The purpose is rather to have a
    quicker or more intuitive way to filter the given completions, especially
    when many completions have a common prefix.

    Fuzzy algorithm is based on this post:
    https://blog.amjith.com/fuzzyfinder-in-10-lines-of-python

    :param completer: A :class:`~.Completer` instance.
    :param WORD: When True, use WORD characters.
    :param pattern: Regex pattern which selects the characters before the
        cursor that are considered for the fuzzy matching.
    :param enable_fuzzy: (bool or `Filter`) Enabled the fuzzy behavior. For
        easily turning fuzzyness on or off according to a certain condition.
    """

    def __init__(
        self,
        completer: Completer,
        WORD: bool = False,
        pattern: str | None = None,
        enable_fuzzy: FilterOrBool = True,
    ) -> None:
        assert pattern is None or pattern.startswith("^")

        self.completer = completer
        self.pattern = pattern
        self.WORD = WORD
        self.pattern = pattern
        self.enable_fuzzy = to_filter(enable_fuzzy)

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        if self.enable_fuzzy():
            return self._get_fuzzy_completions(document, complete_event)
        else:
            return self.completer.get_completions(document, complete_event)

    def _get_pattern(self) -> str:
        if self.pattern:
            return self.pattern
        if self.WORD:
            return r"[^\s]+"
        return "^[a-zA-Z0-9_]*"

    def _get_fuzzy_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        word_before_cursor = document.get_word_before_cursor(
            pattern=re.compile(self._get_pattern())
        )

        # Get completions
        document2 = Document(
            text=document.text[: document.cursor_position - len(word_before_cursor)],
            cursor_position=document.cursor_position - len(word_before_cursor),
        )

        inner_completions = list(
            self.completer.get_completions(document2, complete_event)
        )

        fuzzy_matches: list[_FuzzyMatch] = []

        if word_before_cursor == "":
            # If word before the cursor is an empty string, consider all
            # completions, without filtering everything with an empty regex
            # pattern.
            fuzzy_matches = [_FuzzyMatch(0, 0, compl) for compl in inner_completions]
        else:
            pat = ".*?".join(map(re.escape, word_before_cursor))
            pat = f"(?=({pat}))"  # lookahead regex to manage overlapping matches
            regex = re.compile(pat, re.IGNORECASE)
            for compl in inner_completions:
                matches = list(regex.finditer(compl.text))
                if matches:
                    # Prefer the match, closest to the left, then shortest.
                    best = min(matches, key=lambda m: (m.start(), len(m.group(1))))
                    fuzzy_matches.append(
                        _FuzzyMatch(len(best.group(1)), best.start(), compl)
                    )

            def sort_key(fuzzy_match: _FuzzyMatch) -> tuple[int, int]:
                "Sort by start position, then by the length of the match."
                return fuzzy_match.start_pos, fuzzy_match.match_length

            fuzzy_matches = sorted(fuzzy_matches, key=sort_key)

        for match in fuzzy_matches:
            # Include these completions, but set the correct `display`
            # attribute and `start_position`.
            yield Completion(
                text=match.completion.text,
                start_position=match.completion.start_position
                - len(word_before_cursor),
                # We access to private `_display_meta` attribute, because that one is lazy.
                display_meta=match.completion._display_meta,
                display=self._get_display(match, word_before_cursor),
                style=match.completion.style,
            )

    def _get_display(
        self, fuzzy_match: _FuzzyMatch, word_before_cursor: str
    ) -> AnyFormattedText:
        """
        Generate formatted text for the display label.
        """

        def get_display() -> AnyFormattedText:
            m = fuzzy_match
            word = m.completion.text

            if m.match_length == 0:
                # No highlighting when we have zero length matches (no input text).
                # In this case, use the original display text (which can include
                # additional styling or characters).
                return m.completion.display

            result: StyleAndTextTuples = []

            # Text before match.
            result.append(("class:fuzzymatch.outside", word[: m.start_pos]))

            # The match itself.
            characters = list(word_before_cursor)

            for c in word[m.start_pos : m.start_pos + m.match_length]:
                classname = "class:fuzzymatch.inside"
                if characters and c.lower() == characters[0].lower():
                    classname += ".character"
                    del characters[0]

                result.append((classname, c))

            # Text after match.
            result.append(
                ("class:fuzzymatch.outside", word[m.start_pos + m.match_length :])
            )

            return result

        return get_display()


class FuzzyWordCompleter(Completer):
    """
    Fuzzy completion on a list of words.

    (This is basically a `WordCompleter` wrapped in a `FuzzyCompleter`.)

    :param words: List of words or callable that returns a list of words.
    :param meta_dict: Optional dict mapping words to their meta-information.
    :param WORD: When True, use WORD characters.
    """

    def __init__(
        self,
        words: Sequence[str] | Callable[[], Sequence[str]],
        meta_dict: dict[str, str] | None = None,
        WORD: bool = False,
    ) -> None:
        self.words = words
        self.meta_dict = meta_dict or {}
        self.WORD = WORD

        self.word_completer = WordCompleter(
            words=self.words, WORD=self.WORD, meta_dict=self.meta_dict
        )

        self.fuzzy_completer = FuzzyCompleter(self.word_completer, WORD=self.WORD)

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        return self.fuzzy_completer.get_completions(document, complete_event)


class _FuzzyMatch(NamedTuple):
    match_length: int
    start_pos: int
    completion: Completion



import warnings
from inspect import isawaitable
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List

from .callbacks import SPECS_ALL
from .callbacks import SPECS_SAFE
from .callbacks import CallbacksRegistry
from .callbacks import SpecReference
from .dispatcher import Listener
from .dispatcher import Listeners
from .engines.async_ import AsyncEngine
from .engines.sync import SyncEngine
from .event import BoundEvent
from .event_data import TriggerData
from .exceptions import InvalidDefinition
from .exceptions import InvalidStateValue
from .exceptions import TransitionNotAllowed
from .factory import StateMachineMetaclass
from .graph import iterate_states_and_transitions
from .i18n import _
from .model import Model
from .utils import run_async_from_sync

if TYPE_CHECKING:
    from .event import Event
    from .state import State


class StateMachine(metaclass=StateMachineMetaclass):
    """

    Args:
        model: An optional external object to store state. See :ref:`domain models`.

        state_field (str): The model's field which stores the current state.
            Default: ``state``.

        start_value: An optional start state value if there's no current state assigned
            on the :ref:`domain models`. Default: ``None``.

        rtc (bool): Controls the :ref:`processing model`. Defaults to ``True``
            that corresponds to a **run-to-completion** (RTC) model.

        allow_event_without_transition: If ``False`` when an event does not result in a transition,
            an exception ``TransitionNotAllowed`` will be raised.
            If ``True`` the state machine allows triggering events that may not lead to a state
            :ref:`transition`, including tolerance to unknown :ref:`event` triggers.
            Default: ``False``.

        listeners: An optional list of objects that provies attributes to be used as callbacks.
            See :ref:`listeners` for more details.

    """

    TransitionNotAllowed = TransitionNotAllowed
    """Shortcut for easy exception handling.

    Example::

        try:
            sm.send("an-inexistent-event")
        except sm.TransitionNotAllowed:
            pass
    """

    def __init__(
        self,
        model: Any = None,
        state_field: str = "state",
        start_value: Any = None,
        rtc: bool = True,
        allow_event_without_transition: bool = False,
        listeners: "List[object] | None" = None,
    ):
        self.model = model if model else Model()
        self.state_field = state_field
        self.start_value = start_value
        self.allow_event_without_transition = allow_event_without_transition
        self._callbacks = CallbacksRegistry()
        self._states_for_instance: Dict[State, State] = {}

        self._listeners: Dict[Any, Any] = {}
        """Listeners that provides attributes to be used as callbacks."""

        if self._abstract:
            raise InvalidDefinition(_("There are no states or transitions."))

        self._register_callbacks(listeners or [])

        # Activate the initial state, this only works if the outer scope is sync code.
        # for async code, the user should manually call `await sm.activate_initial_state()`
        # after state machine creation.
        self._engine = self._get_engine(rtc)
        self._engine.start()

    def _get_engine(self, rtc: bool):
        if self._callbacks.has_async_callbacks:
            return AsyncEngine(self, rtc=rtc)

        return SyncEngine(self, rtc=rtc)

    def activate_initial_state(self):
        result = self._engine.activate_initial_state()
        if not isawaitable(result):
            return result
        return run_async_from_sync(result)

    def _processing_loop(self):
        return self._engine.processing_loop()

    def __init_subclass__(cls, strict_states: bool = False):
        cls._strict_states = strict_states
        super().__init_subclass__()

    if TYPE_CHECKING:
        """Makes mypy happy with dynamic created attributes"""

        def __getattr__(self, attribute: str) -> Any: ...

    def __repr__(self):
        current_state_id = self.current_state.id if self.current_state_value else None
        return (
            f"{type(self).__name__}(model={self.model!r}, state_field={self.state_field!r}, "
            f"current_state={current_state_id!r})"
        )

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_rtc"] = self._engine._rtc
        del state["_callbacks"]
        del state["_states_for_instance"]
        del state["_engine"]
        return state

    def __setstate__(self, state):
        listeners = state.pop("_listeners")
        rtc = state.pop("_rtc")
        self.__dict__.update(state)
        self._callbacks = CallbacksRegistry()
        self._states_for_instance: Dict[State, State] = {}

        self._listeners: Dict[Any, Any] = {}

        self._register_callbacks([])
        self.add_listener(*listeners.keys())
        self._engine = self._get_engine(rtc)

    def _get_initial_state(self):
        initial_state_value = self.start_value if self.start_value else self.initial_state.value
        try:
            return self.states_map[initial_state_value]
        except KeyError as err:
            raise InvalidStateValue(initial_state_value) from err

    def bind_events_to(self, *targets):
        """Bind the state machine events to the target objects."""

        for event in self.events:
            trigger = getattr(self, event)
            for target in targets:
                if hasattr(target, event):
                    warnings.warn(
                        f"Attribute '{event}' already exists on {target!r}. Skipping binding.",
                        UserWarning,
                        stacklevel=2,
                    )
                    continue
                setattr(target, event, trigger)

    def _add_listener(self, listeners: "Listeners", allowed_references: SpecReference = SPECS_ALL):
        registry = self._callbacks
        for visited in iterate_states_and_transitions(self.states):
            listeners.resolve(
                visited._specs,
                registry=registry,
                allowed_references=allowed_references,
            )

        return self

    def _register_callbacks(self, listeners: List[object]):
        self._listeners.update({listener: None for listener in listeners})
        self._add_listener(
            Listeners.from_listeners(
                (
                    Listener.from_obj(self, skip_attrs=self._protected_attrs),
                    Listener.from_obj(self.model, skip_attrs={self.state_field}),
                    *(Listener.from_obj(listener) for listener in listeners),
                )
            )
        )

        check_callbacks = self._callbacks.check
        for visited in iterate_states_and_transitions(self.states):
            try:
                check_callbacks(visited._specs)
            except Exception as err:
                raise InvalidDefinition(
                    f"Error on {visited!s} when resolving callbacks: {err}"
                ) from err

        self._callbacks.async_or_sync()

    def add_observer(self, *observers):
        """Add a listener."""
        warnings.warn(
            """Method `add_observer` has been renamed to `add_listener`.""",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.add_listener(*observers)

    def add_listener(self, *listeners):
        """Add a listener.

        Listener are a way to generically add behavior to a :ref:`StateMachine` without changing
        its internal implementation.

        .. seealso::

            :ref:`listeners`.
        """
        self._listeners.update({o: None for o in listeners})
        return self._add_listener(
            Listeners.from_listeners(Listener.from_obj(o) for o in listeners),
            allowed_references=SPECS_SAFE,
        )

    def _repr_html_(self):
        return f'<div class="statemachine">{self._repr_svg_()}</div>'

    def _repr_svg_(self):
        return self._graph().create_svg().decode()

    def _graph(self):
        from .contrib.diagram import DotGraphMachine

        return DotGraphMachine(self).get_graph()

    @property
    def current_state_value(self):
        """Get/Set the current :ref:`state` value.

        This is a low level API, that can be used to assign any valid state value
        completely bypassing all the hooks and validations.
        """
        return getattr(self.model, self.state_field, None)

    @current_state_value.setter
    def current_state_value(self, value):
        if value not in self.states_map:
            raise InvalidStateValue(value)
        setattr(self.model, self.state_field, value)

    @property
    def current_state(self) -> "State":
        """Get/Set the current :ref:`state`.

        This is a low level API, that can be to assign any valid state
        completely bypassing all the hooks and validations.
        """

        try:
            state: State = self.states_map[self.current_state_value].for_instance(
                machine=self,
                cache=self._states_for_instance,
            )
            return state
        except KeyError as err:
            if self.current_state_value is None:
                raise InvalidStateValue(
                    self.current_state_value,
                    _(
                        "There's no current state set. In async code, "
                        "did you activate the initial state? "
                        "(e.g., `await sm.activate_initial_state()`)"
                    ),
                ) from err
            raise InvalidStateValue(self.current_state_value) from err

    @current_state.setter
    def current_state(self, value):
        self.current_state_value = value.value

    @property
    def events(self) -> "List[Event]":
        return [getattr(self, event) for event in self.__class__._events]

    @property
    def allowed_events(self) -> "List[Event]":
        """List of the current allowed events."""
        return [getattr(self, event) for event in self.current_state.transitions.unique_events]

    def _put_nonblocking(self, trigger_data: TriggerData):
        """Put the trigger on the queue without blocking the caller."""
        self._engine.put(trigger_data)

    def send(self, event: str, *args, **kwargs):
        """Send an :ref:`Event` to the state machine.

        .. seealso::

            See: :ref:`triggering events`.

        """
        event_instance: BoundEvent = getattr(
            self, event, BoundEvent(id=event, name=event, _sm=self)
        )
        result = event_instance(*args, **kwargs)
        if not isawaitable(result):
            return result
        return run_async_from_sync(result)



from collections import deque


class ScopedChainDict:
    """Dictionary that can delegate lookups to multiple dicts.

    This provides a basic get/set dict interface that is
    backed by multiple dicts.  Each dict is searched from
    the top most (most recently pushed) scope dict until
    a match is found.

    """
    def __init__(self, *scopes):
        # The scopes are evaluated starting at the top of the stack (the most
        # recently pushed scope via .push_scope()).  If we use a normal list()
        # and push/pop scopes by adding/removing to the end of the list, we'd
        # have to always call reversed(self._scopes) whenever we resolve a key,
        # because the end of the list is the top of the stack.
        # To avoid this, we're using a deque so we can append to the front of
        # the list via .appendleft() in constant time, and iterate over scopes
        # without having to do so with a reversed() call each time.
        self._scopes = deque(scopes)

    def __getitem__(self, key):
        for scope in self._scopes:
            if key in scope:
                return scope[key]
        raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def push_scope(self, scope):
        self._scopes.appendleft(scope)

    def pop_scope(self):
        self._scopes.popleft()



"""Native delayed delivery functionality for Celery workers.

This module provides the DelayedDelivery bootstep which handles setup and configuration
of native delayed delivery functionality when using quorum queues.
"""
from typing import Iterator, List, Optional, Set, Union, ValuesView

from kombu import Connection, Queue
from kombu.transport.native_delayed_delivery import (bind_queue_to_native_delayed_delivery_exchange,
                                                     declare_native_delayed_delivery_exchanges_and_queues)
from kombu.utils.functional import retry_over_time

from celery import Celery, bootsteps
from celery.utils.log import get_logger
from celery.utils.quorum_queues import detect_quorum_queues
from celery.worker.consumer import Consumer, Tasks

__all__ = ('DelayedDelivery',)

logger = get_logger(__name__)


# Default retry settings
RETRY_INTERVAL = 1.0  # seconds between retries
MAX_RETRIES = 3      # maximum number of retries


# Valid queue types for delayed delivery
VALID_QUEUE_TYPES = {'classic', 'quorum'}


class DelayedDelivery(bootsteps.StartStopStep):
    """Bootstep that sets up native delayed delivery functionality.

    This component handles the setup and configuration of native delayed delivery
    for Celery workers. It is automatically included when quorum queues are
    detected in the application configuration.

    Responsibilities:
        - Declaring native delayed delivery exchanges and queues
        - Binding all application queues to the delayed delivery exchanges
        - Handling connection failures gracefully with retries
        - Validating configuration settings
    """

    requires = (Tasks,)

    def include_if(self, c: Consumer) -> bool:
        """Determine if this bootstep should be included.

        Args:
            c: The Celery consumer instance

        Returns:
            bool: True if quorum queues are detected, False otherwise
        """
        return detect_quorum_queues(c.app, c.app.connection_for_write().transport.driver_type)[0]

    def start(self, c: Consumer) -> None:
        """Initialize delayed delivery for all broker URLs.

        Attempts to set up delayed delivery for each broker URL in the configuration.
        Failures are logged but don't prevent attempting remaining URLs.

        Args:
            c: The Celery consumer instance

        Raises:
            ValueError: If configuration validation fails
        """
        app: Celery = c.app

        try:
            self._validate_configuration(app)
        except ValueError as e:
            logger.critical("Configuration validation failed: %s", str(e))
            raise

        broker_urls = self._validate_broker_urls(app.conf.broker_url)
        setup_errors = []

        for broker_url in broker_urls:
            try:
                retry_over_time(
                    self._setup_delayed_delivery,
                    args=(c, broker_url),
                    catch=(ConnectionRefusedError, OSError),
                    errback=self._on_retry,
                    interval_start=RETRY_INTERVAL,
                    max_retries=MAX_RETRIES,
                )
            except Exception as e:
                logger.warning(
                    "Failed to setup delayed delivery for %r: %s",
                    broker_url, str(e)
                )
                setup_errors.append((broker_url, e))

        if len(setup_errors) == len(broker_urls):
            logger.critical(
                "Failed to setup delayed delivery for all broker URLs. "
                "Native delayed delivery will not be available."
            )

    def _setup_delayed_delivery(self, c: Consumer, broker_url: str) -> None:
        """Set up delayed delivery for a specific broker URL.

        Args:
            c: The Celery consumer instance
            broker_url: The broker URL to configure

        Raises:
            ConnectionRefusedError: If connection to the broker fails
            OSError: If there are network-related issues
            Exception: For other unexpected errors during setup
        """
        connection: Connection = c.app.connection_for_write(url=broker_url)
        queue_type = c.app.conf.broker_native_delayed_delivery_queue_type
        logger.debug(
            "Setting up delayed delivery for broker %r with queue type %r",
            broker_url, queue_type
        )

        try:
            declare_native_delayed_delivery_exchanges_and_queues(
                connection,
                queue_type
            )
        except Exception as e:
            logger.warning(
                "Failed to declare exchanges and queues for %r: %s",
                broker_url, str(e)
            )
            raise

        try:
            self._bind_queues(c.app, connection)
        except Exception as e:
            logger.warning(
                "Failed to bind queues for %r: %s",
                broker_url, str(e)
            )
            raise

    def _bind_queues(self, app: Celery, connection: Connection) -> None:
        """Bind all application queues to delayed delivery exchanges.

        Args:
            app: The Celery application instance
            connection: The broker connection to use

        Raises:
            Exception: If queue binding fails
        """
        queues: ValuesView[Queue] = app.amqp.queues.values()
        if not queues:
            logger.warning("No queues found to bind for delayed delivery")
            return

        for queue in queues:
            try:
                logger.debug("Binding queue %r to delayed delivery exchange", queue.name)
                bind_queue_to_native_delayed_delivery_exchange(connection, queue)
            except Exception as e:
                logger.error(
                    "Failed to bind queue %r: %s",
                    queue.name, str(e)
                )
                raise

    def _on_retry(self, exc: Exception, interval_range: Iterator[float], intervals_count: int) -> None:
        """Callback for retry attempts.

        Args:
            exc: The exception that triggered the retry
            interval_range: An iterator which returns the time in seconds to sleep next
            intervals_count: Number of retry attempts so far
        """
        logger.warning(
            "Retrying delayed delivery setup (attempt %d/%d) after error: %s",
            intervals_count + 1, MAX_RETRIES, str(exc)
        )

    def _validate_configuration(self, app: Celery) -> None:
        """Validate all required configuration settings.

        Args:
            app: The Celery application instance

        Raises:
            ValueError: If any configuration is invalid
        """
        # Validate broker URLs
        self._validate_broker_urls(app.conf.broker_url)

        # Validate queue type
        self._validate_queue_type(app.conf.broker_native_delayed_delivery_queue_type)

    def _validate_broker_urls(self, broker_urls: Union[str, List[str]]) -> Set[str]:
        """Validate and split broker URLs.

        Args:
            broker_urls: Broker URLs, either as a semicolon-separated string
                  or as a list of strings

        Returns:
            Set of valid broker URLs

        Raises:
            ValueError: If no valid broker URLs are found or if invalid URLs are provided
        """
        if not broker_urls:
            raise ValueError("broker_url configuration is empty")

        if isinstance(broker_urls, str):
            brokers = broker_urls.split(";")
        elif isinstance(broker_urls, list):
            if not all(isinstance(url, str) for url in broker_urls):
                raise ValueError("All broker URLs must be strings")
            brokers = broker_urls
        else:
            raise ValueError(f"broker_url must be a string or list, got {broker_urls!r}")

        valid_urls = {url for url in brokers}

        if not valid_urls:
            raise ValueError("No valid broker URLs found in configuration")

        return valid_urls

    def _validate_queue_type(self, queue_type: Optional[str]) -> None:
        """Validate the queue type configuration.

        Args:
            queue_type: The configured queue type

        Raises:
            ValueError: If queue type is invalid
        """
        if not queue_type:
            raise ValueError("broker_native_delayed_delivery_queue_type is not configured")

        if queue_type not in VALID_QUEUE_TYPES:
            sorted_types = sorted(VALID_QUEUE_TYPES)
            raise ValueError(
                f"Invalid queue type {queue_type!r}. Must be one of: {', '.join(sorted_types)}"
            )



import sys
from typing import Any, Type

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import Self

from dependency_injector.containers import Container
from dependency_injector.providers import Resource


class Lifespan:
    """A starlette lifespan handler performing container resource initialization and shutdown.

    See https://www.starlette.io/lifespan/ for details.

    Usage:

    .. code-block:: python

        from dependency_injector.containers import DeclarativeContainer
        from dependency_injector.ext.starlette import Lifespan
        from dependency_injector.providers import Factory, Self, Singleton
        from starlette.applications import Starlette

        class Container(DeclarativeContainer):
            __self__ = Self()
            lifespan = Singleton(Lifespan, __self__)
            app = Factory(Starlette, lifespan=lifespan)

    :param container: container instance
    :param resource_type: A :py:class:`~dependency_injector.resources.Resource`
        subclass. Limits the resources to be initialized and shutdown.
    """

    container: Container
    resource_type: Type[Resource[Any]]

    def __init__(
        self,
        container: Container,
        resource_type: Type[Resource[Any]] = Resource,
    ) -> None:
        self.container = container
        self.resource_type = resource_type

    def __call__(self, app: Any) -> Self:
        return self

    async def __aenter__(self) -> None:
        result = self.container.init_resources(self.resource_type)

        if result is not None:
            await result

    async def __aexit__(self, *exc_info: Any) -> None:
        result = self.container.shutdown_resources(self.resource_type)

        if result is not None:
            await result



import importlib
import sys
from urllib.parse import quote
from urllib.request import urlopen

import pydot

from ..statemachine import StateMachine


class DotGraphMachine:
    graph_rankdir = "LR"
    """
    Direction of the graph. Defaults to "LR" (option "TB" for top bottom)
    http://www.graphviz.org/doc/info/attrs.html#d:rankdir
    """

    font_name = "Arial"
    """Graph font face name"""

    state_font_size = "10"
    """State font size in points"""

    state_active_penwidth = 2
    """Active state external line width"""

    state_active_fillcolor = "turquoise"

    transition_font_size = "9"
    """Transition font size in points"""

    def __init__(self, machine: StateMachine):
        self.machine = machine

    def _get_graph(self):
        machine = self.machine
        return pydot.Dot(
            "list",
            graph_type="digraph",
            label=machine.name,
            fontname=self.font_name,
            fontsize=self.state_font_size,
            rankdir=self.graph_rankdir,
        )

    def _initial_node(self):
        node = pydot.Node(
            "i",
            shape="circle",
            style="filled",
            fontsize="1",
            fixedsize="true",
            width=0.2,
            height=0.2,
        )
        node.set_fillcolor("black")
        return node

    def _initial_edge(self):
        return pydot.Edge(
            "i",
            self.machine.initial_state.id,
            label="",
            color="blue",
            fontname=self.font_name,
            fontsize=self.transition_font_size,
        )

    def _actions_getter(self):
        if isinstance(self.machine, StateMachine):

            def getter(grouper) -> str:
                return self.machine._callbacks.str(grouper.key)
        else:

            def getter(grouper) -> str:
                all_names = set(dir(self.machine))
                return ", ".join(
                    str(c) for c in grouper if not c.is_convention or c.func in all_names
                )

        return getter

    def _state_actions(self, state):
        getter = self._actions_getter()

        entry = str(getter(state.enter))
        exit_ = str(getter(state.exit))
        internal = ", ".join(
            f"{transition.event} / {str(getter(transition.on))}"
            for transition in state.transitions
            if transition.internal
        )

        if entry:
            entry = f"entry / {entry}"
        if exit_:
            exit_ = f"exit / {exit_}"

        actions = "\n".join(x for x in [entry, exit_, internal] if x)

        if actions:
            actions = f"\n{actions}"

        return actions

    def _state_as_node(self, state):
        actions = self._state_actions(state)

        node = pydot.Node(
            state.id,
            label=f"{state.name}{actions}",
            shape="rectangle",
            style="rounded, filled",
            fontname=self.font_name,
            fontsize=self.state_font_size,
            peripheries=2 if state.final else 1,
        )
        if state == self.machine.current_state:
            node.set_penwidth(self.state_active_penwidth)
            node.set_fillcolor(self.state_active_fillcolor)
        else:
            node.set_fillcolor("white")
        return node

    def _transition_as_edge(self, transition):
        cond = ", ".join([str(cond) for cond in transition.cond])
        if cond:
            cond = f"\n[{cond}]"
        return pydot.Edge(
            transition.source.id,
            transition.target.id,
            label=f"{transition.event}{cond}",
            color="blue",
            fontname=self.font_name,
            fontsize=self.transition_font_size,
        )

    def get_graph(self):
        graph = self._get_graph()
        graph.add_node(self._initial_node())
        graph.add_edge(self._initial_edge())

        for state in self.machine.states:
            graph.add_node(self._state_as_node(state))
            for transition in state.transitions:
                if transition.internal:
                    continue
                graph.add_edge(self._transition_as_edge(transition))

        return graph

    def __call__(self):
        return self.get_graph()


def quickchart_write_svg(sm: StateMachine, path: str):
    """
    If the default dependency of GraphViz installed locally doesn't work for you. As an option,
    you can generate the image online from the output of the `dot` language,
    using one of the many services available.

    To get the **dot** representation of your state machine is as easy as follows:

    >>> from tests.examples.order_control_machine import OrderControl
    >>> sm = OrderControl()
    >>> print(sm._graph().to_string())
    digraph list {
    ...

    To give you an example, we included this method that will serialize the dot, request the graph
    to https://quickchart.io, and persist the result locally as an ``.svg`` file.


    .. warning::
        Quickchart is an external graph service that supports many formats to generate diagrams.

        By using this method, you should trust http://quickchart.io.

        Please read https://quickchart.io/documentation/faq/ for more information.

    >>> quickchart_write_svg(sm, "docs/images/oc_machine_processing.svg")  # doctest: +SKIP

    """
    dot_representation = sm._graph().to_string()

    url = f"https://quickchart.io/graphviz?graph={quote(dot_representation)}"

    response = urlopen(url)
    data = response.read()

    with open(path, "wb") as f:
        f.write(data)


def import_sm(qualname):
    module_name, class_name = qualname.rsplit(".", 1)
    module = importlib.import_module(module_name)
    smclass = getattr(module, class_name, None)
    if not smclass or not issubclass(smclass, StateMachine):
        raise ValueError(f"{class_name} is not a subclass of StateMachine")

    return smclass


def write_image(qualname, out):
    """
    Given a `qualname`, that is the fully qualified dotted path to a StateMachine
    classes, imports the class and generates a dot graph using the `pydot` lib.
    Writes the graph representation to the filename 'out' that will
    open/create and truncate such file and write on it a representation of
    the graph defined by the statemachine, in the format specified by
    the extension contained in the out path (out.ext).
    """
    smclass = import_sm(qualname)

    graph = DotGraphMachine(smclass).get_graph()
    out_extension = out.rsplit(".", 1)[1]
    graph.write(out, format=out_extension)


def main(argv=None):
    import argparse

    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] <class_path> <out>",
        description="Generate diagrams for StateMachine classes.",
    )
    parser.add_argument(
        "class_path", help="A fully-qualified dotted path to the StateMachine class."
    )
    parser.add_argument(
        "out",
        help="File to generate the image using extension as the output format.",
    )

    args = parser.parse_args(argv)
    write_image(qualname=args.class_path, out=args.out)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



from typing import NamedTuple

from celery import Task


class MapTaskTemplate(NamedTuple):
    template_name: str
    list_attribute: str
    map_index_field: str = "map_index"
    map_value_field: str = "map_value"


CompositeTemplateType = (
    str
    | Task
    | list["CompositeTemplateType"]
    | dict[str, "CompositeTemplateType"]
    | MapTaskTemplate
)
CompositeContentType = (
    str | list["CompositeContentType"] | dict[str, "CompositeContentType"]
)



"""Dependency injector errors."""


class Error(Exception):
    """Base error.

    All dependency injector errors extend this error class.
    """


class NoSuchProviderError(Error, AttributeError):
    """Error that is raised when provider lookup is failed."""


class NonCopyableArgumentError(Error):
    """Error that is raised when provider argument is not deep-copyable."""

    index: int
    keyword: str
    provider: object

    def __init__(self, provider: object, index: int = -1, keyword: str = "") -> None:
        self.provider = provider
        self.index = index
        self.keyword = keyword

    def __str__(self) -> str:
        s = (
            f"keyword argument {self.keyword}"
            if self.keyword
            else f"argument at index {self.index}"
        )
        return f"Couldn't copy {s} for provider {self.provider!r}"



import functools

from django.db import transaction

from celery.app.task import Task


class DjangoTask(Task):
    """
    Extend the base :class:`~celery.app.task.Task` for Django.

    Provide a nicer API to trigger tasks at the end of the DB transaction.
    """

    def delay_on_commit(self, *args, **kwargs) -> None:
        """Call :meth:`~celery.app.task.Task.delay` with Django's ``on_commit()``."""
        transaction.on_commit(functools.partial(self.delay, *args, **kwargs))

    def apply_async_on_commit(self, *args, **kwargs) -> None:
        """Call :meth:`~celery.app.task.Task.apply_async` with Django's ``on_commit()``."""
        transaction.on_commit(functools.partial(self.apply_async, *args, **kwargs))



"""Resources module."""

from abc import ABCMeta, abstractmethod
from typing import Any, ClassVar, Generic, Optional, Tuple, TypeVar

T = TypeVar("T")


class Resource(Generic[T], metaclass=ABCMeta):
    __slots__: ClassVar[Tuple[str, ...]] = ("args", "kwargs", "obj")

    obj: Optional[T]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs
        self.obj = None

    @abstractmethod
    def init(self, *args: Any, **kwargs: Any) -> Optional[T]: ...

    def shutdown(self, resource: Optional[T]) -> None: ...

    def __enter__(self) -> Optional[T]:
        self.obj = obj = self.init(*self.args, **self.kwargs)
        return obj

    def __exit__(self, *exc_info: Any) -> None:
        self.shutdown(self.obj)
        self.obj = None


class AsyncResource(Generic[T], metaclass=ABCMeta):
    __slots__: ClassVar[Tuple[str, ...]] = ("args", "kwargs", "obj")

    obj: Optional[T]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs
        self.obj = None

    @abstractmethod
    async def init(self, *args: Any, **kwargs: Any) -> Optional[T]: ...

    async def shutdown(self, resource: Optional[T]) -> None: ...

    async def __aenter__(self) -> Optional[T]:
        self.obj = obj = await self.init(*self.args, **self.kwargs)
        return obj

    async def __aexit__(self, *exc_info: Any) -> None:
        await self.shutdown(self.obj)
        self.obj = None



"""Worker name utilities."""
from __future__ import annotations

import os
import socket
from functools import partial

from kombu.entity import Exchange, Queue

from .functional import memoize
from .text import simple_format

#: Exchange for worker direct queues.
WORKER_DIRECT_EXCHANGE = Exchange('C.dq2')

#: Format for worker direct queue names.
WORKER_DIRECT_QUEUE_FORMAT = '{hostname}.dq2'

#: Separator for worker node name and hostname.
NODENAME_SEP = '@'

NODENAME_DEFAULT = 'celery'

gethostname = memoize(1, Cache=dict)(socket.gethostname)

__all__ = (
    'worker_direct',
    'gethostname',
    'nodename',
    'anon_nodename',
    'nodesplit',
    'default_nodename',
    'node_format',
    'host_format',
)


def worker_direct(hostname: str | Queue) -> Queue:
    """Return the :class:`kombu.Queue` being a direct route to a worker.

    Arguments:
        hostname (str, ~kombu.Queue): The fully qualified node name of
            a worker (e.g., ``w1@example.com``).  If passed a
            :class:`kombu.Queue` instance it will simply return
            that instead.
    """
    if isinstance(hostname, Queue):
        return hostname
    return Queue(
        WORKER_DIRECT_QUEUE_FORMAT.format(hostname=hostname),
        WORKER_DIRECT_EXCHANGE,
        hostname,
    )


def nodename(name: str, hostname: str) -> str:
    """Create node name from name/hostname pair."""
    return NODENAME_SEP.join((name, hostname))


def anon_nodename(hostname: str | None = None, prefix: str = 'gen') -> str:
    """Return the nodename for this process (not a worker).

    This is used for e.g. the origin task message field.
    """
    return nodename(''.join([prefix, str(os.getpid())]), hostname or gethostname())


def nodesplit(name: str) -> tuple[None, str] | list[str]:
    """Split node name into tuple of name/hostname."""
    parts = name.split(NODENAME_SEP, 1)
    if len(parts) == 1:
        return None, parts[0]
    return parts


def default_nodename(hostname: str) -> str:
    """Return the default nodename for this process."""
    name, host = nodesplit(hostname or '')
    return nodename(name or NODENAME_DEFAULT, host or gethostname())


def node_format(s: str, name: str, **extra: dict) -> str:
    """Format worker node name (name@host.com)."""
    shortname, host = nodesplit(name)
    return host_format(s, host, shortname or NODENAME_DEFAULT, p=name, **extra)


def _fmt_process_index(prefix: str = '', default: str = '0') -> str:
    from .log import current_process_index

    index = current_process_index()
    return f'{prefix}{index}' if index else default


_fmt_process_index_with_prefix = partial(_fmt_process_index, '-', '')


def host_format(s: str, host: str | None = None, name: str | None = None, **extra: dict) -> str:
    """Format host %x abbreviations."""
    host = host or gethostname()
    hname, _, domain = host.partition('.')
    name = name or hname
    keys = dict(
        {
            'h': host,
            'n': name,
            'd': domain,
            'i': _fmt_process_index,
            'I': _fmt_process_index_with_prefix,
        },
        **extra,
    )
    return simple_format(s, keys)



import hashlib
import typing

from loguru import logger
from statemachine import State
from statemachine.event_data import EventData

from genie_flow.genie import StateType, DialoguePersistence
from genie_flow.model.template import CompositeTemplateType

if typing.TYPE_CHECKING:
    from genie_flow.celery import CeleryManager


_DIALOGUE_PERSISTENCE_MAP: dict[StateType, dict[StateType, DialoguePersistence]] = {
    StateType.USER: {
        StateType.USER: DialoguePersistence.RENDERED,
        StateType.INVOKER: DialoguePersistence.RAW,
    },
    StateType.INVOKER: {
        StateType.USER: DialoguePersistence.RENDERED,
        StateType.INVOKER: DialoguePersistence.NONE,
    }
}

class TransitionManager:

    def __init__(self, celery_manager: "CeleryManager"):
        self.celery_manager = celery_manager

    def _determine_transition_type(self, event_data: EventData) -> tuple[StateType, StateType]:
        def determine(state: State) -> StateType:
            state_template: CompositeTemplateType = event_data.machine.get_template_for_state(state)
            return (
                StateType.INVOKER
                if self.celery_manager.genie_environment.has_invoker(state_template)
                else StateType.USER
            )

        return determine(event_data.source), determine(event_data.target)

    def before_transition(self, event_data: EventData):
        """
        This hook determines how the transition will be conducted. It will
        set the property `transition_type` to a tuple containing the source type
        and the destination type. This hook also determines if and how the event
        argument should be stored as part of the dialogue. The property
        `dialogue_persistence` is set to "NONE", "RAW" or "RENDERED". Finally, this
        hook also sets the `actor_input` property to the first argument that was
        passed with the triggering event.

        :param event_data: The event data object provided by the state machine
        """
        logger.debug(
            "starting transition for session {session_id}, "
            "to state {state_id} with event {event_id}",
            session_id=event_data.machine.model.session_id,
            state_id=event_data.target.id,
            event_id=event_data.event,
        )

        source_type, target_type = self._determine_transition_type(event_data)
        event_data.machine.model.source_type = source_type
        event_data.machine.model.target_type = target_type
        event_data.machine.model.actor = source_type.as_actor
        logger.debug(
            "determined transition type for session {session_id}, "
            "to state {state_id} with event {event_id} "
            "to be from {source_type} to {target_type}, with actor {actor}",
            session_id=event_data.machine.model.session_id,
            state_id=event_data.target.id,
            event_id=event_data.event,
            source_type=source_type.name,
            target_type=target_type.name,
            actor=event_data.machine.model.actor,
        )

        dialogue_persistence = (
            _DIALOGUE_PERSISTENCE_MAP[source_type][target_type]
        )
        event_data.machine.model.dialogue_persistence = dialogue_persistence
        logger.debug(
            "determined dialogue persistence for session {session_id}, "
            "to state {state_id} with event {event_id} "
            "to be {dialogue_persistence}",
            session_id=event_data.machine.model.session_id,
            state_id=event_data.target.id,
            event_id=event_data.event,
            dialogue_persistence=dialogue_persistence.name,
        )

        actor_input : str = (
            event_data.args[0]
            if event_data.args is not None and len(event_data.args) > 0
            else None
        )
        logger.debug("set actor input to '{actor_input}'", actor_input=actor_input)
        logger.info(
            "set actor input to string of md5 hash {actor_input_hash}",
            actor_input_hash=(
                hashlib.md5(actor_input.encode("utf-8")).hexdigest()
                if actor_input is not None
                else None
            ),
        )
        event_data.machine.model.actor_input = actor_input

    def after_transition(self, event_data: EventData):
        logger.debug(
            "after transition for session {session_id}, "
            "to state {state_id} with event {event_id} "
            "and dialogue persistence: {dialogue_persistence}",
            session_id=event_data.machine.model.session_id,
            state_id=event_data.target.id,
            event_id=event_data.event,
            dialogue_persistence=event_data.machine.model.dialogue_persistence.name,
        )

        if event_data.machine.model.target_type == StateType.INVOKER:
            logger.info(
                "enqueueing task for session {session_id}, "
                "to state {state_id} with event {event_id}",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
            )
            self.celery_manager.enqueue_task(
                event_data.machine,
                event_data.machine.model,
                event_data.target,
            )

        if event_data.machine.model.dialogue_persistence == DialoguePersistence.NONE:
            logger.info(
                "not recording dialogue for session {session_id}, "
                "to state {state_id} with event {event_id}",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
            )
            return

        if event_data.machine.model.dialogue_persistence == DialoguePersistence.RENDERED:
            logger.info(
                "rendering template for session {session_id}, "
                "to state {state_id} with event {event_id}",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
            )
            target_template_path = event_data.machine.get_template_for_state(
                event_data.machine.current_state,
            )
            actor_input = self.celery_manager.genie_environment.render_template(
                template_path=target_template_path,
                data_context=event_data.machine.model.render_data,
            )
            logger.debug(
                "recording rendered output for session {session_id}, "
                "to state {state_id} with event {event_id} "
                "as: '{actor_input}'",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
                actor_input=(
                    f"{actor_input[:50]}..."
                    if len(actor_input) > 50 else actor_input
                ),
            )
            event_data.machine.model.actor_input = actor_input
        else:
            logger.debug(
                "recording raw output for session {session_id}, "
                "to state {state_id} with event {event_id} "
                "as: '{actor_input}'",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
                actor_input=(
                    f"{event_data.machine.model.actor_input[:50]}..."
                    if len(event_data.machine.model.actor_input) > 50
                    else event_data.machine.model.actor_input
                ),
            )
            logger.info(
                "adding raw actor input to dialogue for session {session_id}, "
                "to state {state_id} with event {event_id}",
                session_id=event_data.machine.model.session_id,
                state_id=event_data.target.id,
                event_id=event_data.event,
            )

        event_data.machine.model.record_dialogue_element()
